#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.yaml."""

from breezy.tests import (
    TestCase,
    TestCaseInTempDir,
)
from ruamel.yaml.compat import ordereddict

from lintian_brush.yaml import (
    YamlUpdater,
    update_ordered_dict,
)


class YamlUpdaterTests(TestCaseInTempDir):
    def test_new(self):
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Somekey"] = "Somevalue"
        self.assertFileEqual("---\nSomekey: Somevalue\n", "newfile.yaml")

    def test_update(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
Origkey: origvalue
Somekey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Somekey"] = "Somevalue"
        self.assertFileEqual(
            """\
Origkey: origvalue
Somekey: Somevalue
""",
            "newfile.yaml",
        )

    def test_update_with_spaces(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
Origkey : origvalue
Somekey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Somekey"] = "Somevalue"
        self.assertFileEqual(
            """\
Origkey : origvalue
Somekey: Somevalue
""",
            "newfile.yaml",
        )

    def test_delete(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
Origkey: origvalue
Somekey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            del editor.code["Origkey"]
            del editor.code["Somekey"]
        self.assertPathDoesNotExist("newfile.yaml")

    def test_no_change(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
Origkey: origvalue
Somekey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml"):
            pass
        self.assertFileEqual(
            """\
Origkey: origvalue
Somekey: origvalue
""",
            "newfile.yaml",
        )

    def test_preserve_header(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
---
Origkey: origvalue
Somekey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Newkey"] = "newvalue"
        self.assertFileEqual(
            """\
---
Origkey: origvalue
Somekey: origvalue
Newkey: newvalue
""",
            "newfile.yaml",
        )

    def test_interrupted_line(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
---
Origkey: origvalue
Somekey: origvalue"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Newkey"] = "newvalue"
        self.assertFileEqual(
            """\
---
Origkey: origvalue
Somekey: origvalue
Newkey: newvalue
""",
            "newfile.yaml",
        )

    def test_replace_only_line(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
Origkey: origvalue
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            del editor.code["Origkey"]
            editor.code["Newkey"] = "newvalue"
        self.assertFileEqual(
            """\
Newkey: newvalue
""",
            "newfile.yaml",
        )

    def test_json(self):
        with open("newfile.yaml", "w") as f:
            f.write(
                """\
{
  "Archive": "GitHub",
  "Bug-Database": "https://github.com/fabiang/sasl/issues",
  "Bug-Submit": "https://github.com/fabiang/sasl/issues/new",
  "Documentation": "https://github.com/fabiang/sasl#usage",
  "Name": "The PHP SASL2 Authentification Library",
}
"""
            )
        with YamlUpdater("newfile.yaml") as editor:
            editor.code["Repository"] = "git@github.com:fabiang/sasl.git"
        self.assertFileEqual(
            """\
{
  "Archive": "GitHub",
  "Bug-Database": "https://github.com/fabiang/sasl/issues",
  "Bug-Submit": "https://github.com/fabiang/sasl/issues/new",
  "Documentation": "https://github.com/fabiang/sasl#usage",
  "Name": "The PHP SASL2 Authentification Library",
  "Repository": "git@github.com:fabiang/sasl.git"
}
""",
            "newfile.yaml",
        )


class UpdateOrderedDict(TestCase):
    def setUp(self):
        super().setUp()
        self._od = ordereddict()

    def test_empty(self):
        update_ordered_dict(self._od, [("Contact", "Foo"), ("Blah", "blah")])
        self.assertEqual(
            ordereddict([("Blah", "blah"), ("Contact", "Foo")]), self._od)

    def test_modify(self):
        self._od["Contact"] = "Bar"
        self._od["ZZ"] = "z"
        update_ordered_dict(self._od, [("Contact", "Foo"), ("Blah", "blah")])
        self.assertEqual(
            ordereddict(
                [
                    ("Blah", "blah"),
                    ("Contact", "Foo"),
                    ("ZZ", "z"),
                ]
            ),
            self._od,
        )

    def test_insert_before(self):
        self._od["Contact"] = "Bar"
        self._od["Bar"] = "Bar"
        update_ordered_dict(self._od, [("Daar", "blah")])
        self.assertEqual(
            ordereddict(
                [
                    ("Contact", "Bar"),
                    ("Bar", "Bar"),
                    ("Daar", "blah"),
                ]
            ),
            self._od,
        )

    def test_csb(self):
        from upstream_ontologist import upstream_metadata_sort_key

        self._od["Registry"] = [
            ordereddict([("Name", "OMICtools"), ("Entry", "OMICS_09827")]),
            ordereddict([("Name", "bio.tools"), ("Entry", "NA")]),
        ]
        self._od["Repository"] = "https://github.com/csb-toolbox/CSB"
        update_ordered_dict(
            self._od,
            [
                ("Bug-Database", "https://github.com/csb-toolbox/CSB/issues"),
                ("Bug-Submit",
                 "https://github.com/csb-toolbox/CSB/issues/new"),
                ("Repository",
                 "https://github.com/csb-toolbox/CSB/issues.git"),
            ],
            key=upstream_metadata_sort_key,
        )
        self.assertEqual(
            ordereddict(
                [
                    ("Bug-Database",
                     "https://github.com/csb-toolbox/CSB/issues"),
                    ("Bug-Submit",
                     "https://github.com/csb-toolbox/CSB/issues/new"),
                    ("Registry", [
                        ordereddict(
                            [("Name", "OMICtools"), ("Entry", "OMICS_09827")]
                        ),
                        ordereddict([("Name", "bio.tools"), ("Entry", "NA")]),
                    ]),
                    ("Repository",
                     "https://github.com/csb-toolbox/CSB/issues.git"),
                ]
            ),
            self._od,
        )



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import report_result

if not os.path.exists('debian/rules'):
    sys.exit(0)

os.chmod('debian/rules', 0o755)

report_result(
    'Mark debian/rules as executable.',
    certainty='certain')



#!/usr/bin/python3

import os
import re

from debmutate.reformatting import GeneratedFile, check_generated_file

from lintian_brush.fixer import LintianIssue, report_result

trailing_whitespace_re = re.compile(b'\\s*\n$')
trailing_space_re = re.compile(b'[ ]*\n$')


def strip_whitespace(line: bytes, strip_tabs=True) -> bytes:
    if strip_tabs:
        pat = trailing_whitespace_re
    else:
        pat = trailing_space_re
    return re.sub(pat, b'\n', line)


def file_strip_whitespace(
        path: str, strip_tabs: bool = True,
        strip_trailing_empty_lines: bool = True,
        delete_new_empty_line: bool = False) -> bool:
    newlines = []
    changed = False
    try:
        with open(path, 'rb') as f:
            for lineno, line in enumerate(f, 1):
                newline = strip_whitespace(line, strip_tabs=strip_tabs)
                if newline != line:
                    issue = LintianIssue(
                        'source', 'trailing-whitespace',
                        info='%s (line %d)' % (path, lineno))
                    if issue.should_fix():
                        issue.report_fixed()
                        changed = True
                        if newline == b'\n' and delete_new_empty_line:
                            continue
                newlines.append(newline)
    except FileNotFoundError:
        return False
    if strip_trailing_empty_lines:
        while newlines and newlines[-1] == b'\n':
            issue = LintianIssue(
                'source', 'trailing-whitespace',
                info='%s (line %d)' % (path, len(newlines) - 1))
            if issue.should_fix():
                issue.report_fixed()
                changed = True
                newlines.pop(-1)
            else:
                break
    if changed:
        with open(path, 'wb') as f:
            f.writelines(newlines)
        return True
    return False


file_strip_whitespace('debian/changelog', strip_tabs=True)
file_strip_whitespace('debian/rules', strip_tabs=False)

try:
    check_generated_file('debian/control')
except GeneratedFile:
    changed = False
    for entry in os.scandir('debian'):
        if not entry.name.startswith('control.') or entry.name.endswith('~'):
            continue
        if entry.name.endswith('.m4'):
            continue
        if file_strip_whitespace(
                entry.path, strip_tabs=True,
                delete_new_empty_line=True):
            changed = True
    if changed:
        file_strip_whitespace(
            'debian/control', strip_tabs=True, delete_new_empty_line=True)
else:
    file_strip_whitespace(
        'debian/control', strip_tabs=True, delete_new_empty_line=True)

report_result("Trim trailing whitespace.")



#!/usr/bin/python3

import re

from lintian_brush import available_lintian_fixers

with open('README.md') as f:
    contents = f.read()

fixers = available_lintian_fixers()

tags = set()
for fixer in fixers:
    tags.update(fixer.lintian_tags)
replacement_text = ''.join(['* %s\n' % tag for tag in sorted(tags)])

with open('README.md', 'w') as f:
    # TODO(jelmer): Use better sentinels, just in case somebody changes
    # the current ones?
    f.write(re.sub(
        r'(subset of the issues:\n\n).*(\n\.\. _writing-fixers:\n)',
        '\\1' + replacement_text + '\\2', contents,
        flags=re.MULTILINE | re.DOTALL))



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Debianize a package."""

__all__ = [
    'NoBuildToolsFound',
    'debianize',
    ]

import contextlib
import errno
import json
import logging
import os
import shutil
import subprocess
from dataclasses import dataclass, field
from functools import partial
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict, List, Optional, Tuple, cast
from urllib.parse import urlparse

from breezy import osutils
from breezy.branch import Branch
from breezy.commit import NullCommitReporter, PointlessCommit
from breezy.controldir import ControlDir
from breezy.errors import (
    AlreadyBranchError,
    NoSuchRevision,
    NotBranchError,
)
from breezy.plugins.debian import default_orig_dir
from breezy.plugins.debian.directory import vcs_git_url_to_bzr_url
from breezy.plugins.debian.import_dsc import UpstreamAlreadyImported
from breezy.plugins.debian.merge_upstream import (
    do_import,
    get_existing_imported_upstream_revids,
    get_tarballs,
)
from breezy.plugins.debian.upstream import (
    PackageVersionNotPresent,
    UpstreamSource,
)
from breezy.plugins.debian.upstream.branch import (
    DistCommandFailed,
    UpstreamBranchSource,
    run_dist_command,
    upstream_version_add_revision,
)
from breezy.plugins.debian.upstream.pristinetar import (
    BasePristineTarSource,
    get_pristine_tar_source,
)
from breezy.revision import NULL_REVISION, RevisionID
from breezy.transport import FileExists
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.workspace import (
    WorkspaceDirty,
    check_clean_tree,
    reset_tree,
)
from buildlog_consultant.common import (
    SetuptoolScmVersionIssue,
    VcsControlDirectoryNeeded,
)
from debmutate.control import (
    ControlEditor,
    ensure_exact_version,
    ensure_relation,
    ensure_some_version,
)
from debmutate.vcs import unsplit_vcs_url
from debmutate.versions import (
    debianize_upstream_version,
)
from ognibuild import DetailedFailure, UnidentifiedError
from ognibuild.buildlog import problem_to_upstream_requirement
from ognibuild.buildsystem import (
    BuildSystem,
    NoBuildToolsFound,
    get_buildsystem,
)
from ognibuild.debian.apt import AptManager
from ognibuild.debian.fix_build import (
    DetailedDebianBuildFailure,
    UnidentifiedDebianBuildError,
    build_incrementally,
    default_fixers,
)
from ognibuild.debian.upstream_deps import get_project_wide_deps
from ognibuild.dist import (  # noqa: F401
    DistNoTarball,
)
from ognibuild.dist import (
    create_dist as ogni_create_dist,
)
from ognibuild.fix_build import BuildFixer, iterate_with_build_fixers
from ognibuild.requirements import (
    Requirement,
)
from ognibuild.resolver.apt import AptRequirement
from ognibuild.session import Session, SessionSetupFailure
from ognibuild.session.plain import PlainSession
from ognibuild.session.schroot import SchrootSession
from ognibuild.upstream import (
    find_apt_upstream,
    find_upstream,
    go_base_name,
    load_crate_info,
)
from ognibuild.vcs import dupe_vcs_tree
from upstream_ontologist.debian import (
    upstream_name_to_debian_source_name as source_name_from_upstream_name,
)
from upstream_ontologist.debian import (
    upstream_version_to_debian_upstream_version as debian_upstream_version,
)
from upstream_ontologist.debian import (
    valid_debian_package_name,
)
from upstream_ontologist.guess import (
    UpstreamDatum,
    guess_upstream_info,
    summarize_upstream_metadata,
)

from debian.changelog import Changelog, Version, format_date, get_maintainer
from debian.deb822 import PkgRelation

from . import (
    _lintian_brush_rs,
    available_lintian_fixers,
    get_committer,
    run_lintian_fixers,
)
from .debbugs import find_archived_wnpp_bugs, find_wnpp_bugs
from .debhelper import (
    maximum_debhelper_compat_version,
)
from .debhelper import (
    write_rules_template as write_debhelper_rules_template,
)
from .publish import (
    NoVcsLocation,
    update_official_vcs,
)
from .standards_version import latest_standards_version
from .svp import svp_enabled

Kickstarter = Callable[[WorkingTree, str], None]


class BuildSystemProcessError(Exception):
    """Error processing buildsystem-specific part of debianization."""

    def __init__(self, buildsystem, message, inner=None):
        self.buildsystem = buildsystem
        self.message = message
        self.inner = inner


class DebianizedPackageRequirementMismatch(Exception):
    """Debianized package does not satisfy requirement."""

    def __init__(self, requirement, control, version, upstream_branch):
        self.requirement = requirement
        self.control = control
        self.version = version
        self.upstream_branch = upstream_branch


class SourcePackageNameInvalid(Exception):
    """Source package name is invalid."""

    def __init__(self, source):
        self.source = source


class SourceNameUnknown(Exception):
    """Unable to determine source name."""

    def __init__(self, upstream_name):
        self.upstream_name = upstream_name


class NoUpstreamReleases(Exception):
    """No upstream releases were found."""

    def __init__(self, upstream_source, name):
        self.upstream_source = upstream_source
        self.name = name


def write_changelog_template(
        path, source_name, version, author=None, wnpp_bugs=None):
    if author is None:
        author = get_maintainer()
    if wnpp_bugs:
        closes = " Closes: " + ", ".join(
            [("#%d" % (bug,)) for bug, kind in wnpp_bugs])
    else:
        closes = ""
    cl = Changelog()
    cl.new_block(
        package=source_name,
        version=version,
        distributions="UNRELEASED",
        urgency="low",
        changes=["", "  * Initial release." + closes, ""],
        author="{} <{}>".format(*author),
        date=format_date(),
    )
    with open(path, "w") as f:
        f.write(cl.__str__().strip("\n") + "\n")


MINIMUM_CERTAINTY = "possible"  # For now..


def versions_dict():
    import buildlog_consultant
    import debmutate
    import ognibuild
    import upstream_ontologist

    import debian
    import lintian_brush
    return {
        'lintian-brush': lintian_brush.version_string,
        'debmutate': debmutate.version_string,
        'debian': debian.__version__,
        'ognibuild': ognibuild.version_string,
        'buildlog_consultant': buildlog_consultant.version_string,
        'upstream_ontologist': upstream_ontologist.version_string,
    }


class DebianDirectoryExists(Exception):
    """A Debian Directory already exists."""

    def __init__(self, path):
        self.path = path


def go_import_path_from_repo(repo_url):
    parsed_url = urlparse(repo_url)
    p = parsed_url.hostname + parsed_url.path.rstrip('/')
    if p.endswith(".git"):
        p = p[:-4]
    return p


def enable_dh_addon(source, addon):
    source["Build-Depends"] = ensure_some_version(
        source["Build-Depends"], "dh-sequence-%s" % addon
    )


def setup_debhelper(
        wt, debian_path, source, compat_release, addons=None, env=None,
        buildsystem=None):
    source["Build-Depends"] = ensure_exact_version(
            source.get("Build-Depends", ""),
            "debhelper-compat",
            str(maximum_debhelper_compat_version(compat_release)))
    for addon in addons or []:
        enable_dh_addon(source, addon)
    write_debhelper_rules_template(
        wt.abspath(os.path.join(debian_path, "rules")),
        buildsystem=buildsystem,
        env=(env or {}),
    )


def default_create_dist(
        session, tree, package, version, target_dir, subpath=""):
    try:
        with session:
            try:
                return ogni_create_dist(
                    session, tree, target_dir,
                    include_controldir=False,
                    subdir=(package or "package"),
                    version=version,
                    subpath=subpath)
            except DetailedFailure as e:
                if isinstance(
                        e.error,
                        (VcsControlDirectoryNeeded, SetuptoolScmVersionIssue)):
                    return ogni_create_dist(
                        session, tree, target_dir,
                        include_controldir=True,
                        subdir=(package or "package"),
                        version=version, subpath=subpath)
                else:
                    raise
    except NoBuildToolsFound:
        logging.info(
            "No build tools found, falling back to simple export.")
        return None
    except NotImplementedError:
        logging.info(
            "Build system does not support dist, falling back "
            "to export.")
        return None
    except SessionSetupFailure as exc:
        raise DistCommandFailed(str(exc), 'session-setup-failure') from exc
    except DistNoTarball as e:
        logging.info("Build system did not create a tarball: %s", e)
        return None
    except DetailedFailure as exc:
        raise DistCommandFailed(str(exc), exc.error) from exc
    except UnidentifiedError as exc:
        raise DistCommandFailed(str(exc)) from exc


def import_upstream_version_from_dist(
        wt, subpath, upstream_source, source_name, upstream_version,
        session: Session) -> Tuple[
            Dict[Optional[str], Tuple[RevisionID, str]],
            Dict[Optional[str], RevisionID], str]:
    orig_dir = os.path.abspath(default_orig_dir)

    tag_names = {}
    with TemporaryDirectory() as target_dir:
        locations = upstream_source.fetch_tarballs(
            source_name, upstream_version, target_dir, components=[None])
        if source_name is None:
            source_name = os.path.basename(locations[0]).split('_')[0]
        try:
            tarball_filenames = get_tarballs(
                orig_dir,
                wt, source_name, upstream_version, locations)
        except FileExists as e:
            logging.warning(
                'Tarball %s exists, reusing existing file.', e.path)
            tarball_filenames = [os.path.join(orig_dir, e.path)]
        upstream_revisions = upstream_source\
            .version_as_revisions(source_name, upstream_version)
        files_excluded = None
        try:
            imported_revids = do_import(
                wt, subpath, tarball_filenames, source_name, upstream_version,
                current_version=None,
                upstream_branch=upstream_source.upstream_branch,
                upstream_revisions=upstream_revisions,
                merge_type=None, files_excluded=files_excluded)
        except UpstreamAlreadyImported as e:
            logging.warning(
                'Upstream release %s already imported.',
                e.version)  # type: ignore
            imported_revids = get_existing_imported_upstream_revids(
                upstream_source, source_name, upstream_version)
        pristine_revids = {}
        for (component, tag_name, revid,
             _pristine_tar_imported, subpath) in imported_revids:
            pristine_revids[component] = (revid, subpath)
            tag_names[component] = tag_name

    upstream_branch_name = "upstream"
    try:
        branch = wt.controldir.create_branch(upstream_branch_name)
    except AlreadyBranchError:
        logging.info("Upstream branch already exists; not creating.")
    else:
        branch.generate_revision_history(pristine_revids[None])
        logging.info("Created upstream branch.")

    return pristine_revids, tag_names, upstream_branch_name


class ResetOnFailure:

    def __init__(self, wt, subpath=None):
        self.wt = wt
        self.subpath = subpath

    def __enter__(self):
        check_clean_tree(self.wt, self.wt.basis_tree(), self.subpath)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            reset_tree(self.wt, self.wt.basis_tree(), self.subpath)
        return False


def process_setup_py(es, session, wt, subpath, debian_path, upstream_version,
                     metadata, compat_release, buildsystem,
                     buildsystem_subpath, kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    source["Rules-Requires-Root"] = "no"
    source["Standards-Version"] = latest_standards_version()
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release,
        addons=["python3"],
        buildsystem="pybuild")
    source["Testsuite"] = "autopkgtest-pkg-python"
    # TODO(jelmer): check whether project supports python 3
    source["Build-Depends"] = ensure_some_version(
        source["Build-Depends"], "python3-all")
    # TODO(Jelmer): Convert pyproject's build-system.requirements to Python
    # deps
    if buildsystem.build_backend in ("flit.build_api", "flit_core.build_api"):
        source["Build-Depends"] = ensure_some_version(
            source["Build-Depends"], "flit")
        source["Build-Depends"] = ensure_some_version(
            source["Build-Depends"], "python3-toml")
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    # We're going to be running the testsuite as part of the build,
    # so import the test dependencies too.
    import_build_deps(source, test_deps)
    upstream_name = metadata['Name']
    if upstream_name.startswith('python-'):
        upstream_name = upstream_name[len('python-'):]
    source['Source'] = "python-%s" % upstream_name.lower().replace('_', '-')
    control.add_binary({
            "Package": "python3-%s" % upstream_name.lower().replace('_', '-'),
            "Depends": "${python3:Depends}",
            "Architecture": "all",
        }
    )
    return control


def process_maven(es, session, wt, subpath, debian_path, upstream_version,
                  metadata, compat_release, buildsystem, buildsystem_subpath,
                  kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(
        ControlEditor.create(wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    source["Rules-Requires-Root"] = "no"
    source["Standards-Version"] = latest_standards_version()
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release,
        buildsystem="maven")
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    upstream_name = metadata['Name']
    source['Source'] = upstream_name
    control.add_binary({
            "Package": "lib%s-java" % upstream_name,
            "Depends": "${java:Depends}",
            "Architecture": "all",
        }
    )
    return control


def process_npm(es, session, wt, subpath, debian_path, upstream_version,
                metadata, compat_release, buildsystem, buildsystem_subpath,
                kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(
        ControlEditor.create(wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release, addons=["nodejs"])
    upstream_name = (
        metadata['Name'].strip('@').replace('/', '-')
        .replace('_', '-').replace('@', '').lower())
    source['Source'] = "node-%s" % upstream_name
    source["Rules-Requires-Root"] = "no"
    source["Standards-Version"] = latest_standards_version()
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    control.add_binary(
        {"Package": "node-%s" % upstream_name, "Architecture": "all"})
    source["Testsuite"] = "autopkgtest-pkg-nodejs"
    return control


def perl_package_name(upstream_name):
    if upstream_name.startswith('lib'):
        upstream_name = upstream_name[len('lib'):]
    return ("lib%s-perl" % upstream_name.replace('::', '-')
            .replace('_', '').lower())


def process_dist_zilla(es, session, wt, subpath, debian_path, upstream_version,
                       metadata, compat_release, buildsystem,
                       buildsystem_subpath, kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(
        ControlEditor.create(wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    upstream_name = metadata['Name']
    source['Source'] = perl_package_name(upstream_name)
    source["Rules-Requires-Root"] = "no"
    source['Testsuite'] = 'autopkgtest-pkg-perl'
    source["Standards-Version"] = latest_standards_version()
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release,
        addons=["dist-zilla"])
    control.add_binary(
        {"Package": source['Source'],
         "Depends": "${perl:Depends}",
         "Architecture": "all"
         })
    return control


def process_makefile_pl(es, session, wt, subpath, debian_path,
                        upstream_version, metadata, compat_release,
                        buildsystem, buildsystem_subpath, kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    upstream_name = metadata['Name']
    source['Source'] = perl_package_name(upstream_name)
    source["Rules-Requires-Root"] = "no"
    source['Testsuite'] = 'autopkgtest-pkg-perl'
    source["Standards-Version"] = latest_standards_version()
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(wt, debian_path, source, compat_release=compat_release)
    control.add_binary(
        {"Package": source['Source'],
         "Depends": "${perl:Depends}",
         "Architecture": "all"
         })
    return control


def process_perl_build_tiny(es, session, wt, subpath, debian_path,
                            upstream_version, metadata, compat_release,
                            buildsystem, buildsystem_subpath,
                            kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    upstream_name = metadata['Name']
    source['Source'] = perl_package_name(upstream_name)
    source["Rules-Requires-Root"] = "no"
    source['Testsuite'] = 'autopkgtest-pkg-perl'
    source["Standards-Version"] = latest_standards_version()
    source["Build-Depends"] = "libmodule-build-perl"
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(wt, debian_path, source, compat_release=compat_release)
    control.add_binary(
        {"Package": source['Source'],
         "Depends": "${perl:Depends}",
         "Architecture": "all"
         })
    return control


def process_golang(es, session, wt, subpath, debian_path, upstream_version,
                   metadata, compat_release, buildsystem, buildsystem_subpath,
                   kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    source["Rules-Requires-Root"] = "no"
    source["Standards-Version"] = latest_standards_version()
    source["XS-Go-Import-Path"] = go_import_path_from_repo(
        metadata["Repository"]
    )
    if "Repository-Browse" in metadata:
        source["Homepage"] = metadata["Repository-Browse"]
    source["Section"] = "devel"
    parsed_url = urlparse(metadata["Repository"])
    godebname = go_base_name(parsed_url.hostname + parsed_url.path)
    source['Source'] = "golang-%s" % godebname
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    source["Testsuite"] = "autopkgtest-pkg-go"
    dh_env = {}
    if os.path.isdir("examples"):
        dh_env["DH_GOLANG_EXCLUDES"] = "examples/"
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release,
        addons=["golang"],
        buildsystem="golang",
        env=dh_env)
    # TODO(jelmer): Add --builddirectory=_build to dh arguments
    control.add_binary({
        "Package": "golang-%s-dev" % godebname,
        "Architecture": "all",
        "Multi-Arch": "foreign",
        })
    return control


def process_r(es, session, wt, subpath, debian_path,
              upstream_version, metadata, compat_release, buildsystem,
              buildsystem_subpath, kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source

    if metadata.get('Archive') == 'CRAN':
        archive = 'cran'
    elif metadata.get('Archive') == 'Bioconductor':
        archive = 'bioc'
    else:
        archive = 'other'

    source["Source"] = "r-{}-{}".format(archive, metadata['Name'].lower())
    source["Rules-Requires-Root"] = "no"
    source["Build-Depends"] = "dh-r, r-base-dev"
    source["Standards-Version"] = latest_standards_version()
    source["Testsuite"] = "autopkgtest-pkg-r"
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(
        wt, debian_path, source, compat_release=compat_release,
        buildsystem="R")
    # For now, just assume a single binary package that is
    # architecture-dependent.
    control.add_binary({
        "Package": "r-{}-{}".format(archive, metadata['Name'].lower()),
        "Architecture": 'any',
        'Depends': '${R:Depends}, ${shlibs:Depends}, ${misc:Depends}',
        'Recommends': '${R:Recommends}',
        'Suggests': '${R:Suggests}',
        })
    return control


def process_octave(es, session, wt, subpath, debian_path, metadata,
                   upstream_version, compat_release, buildsystem,
                   buildsystem_subpath, kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source

    source["Source"] = "octave-%s" % metadata['Name'].lower()
    source["Rules-Requires-Root"] = "no"
    source["Build-Depends"] = "dh-octave"
    source["Standards-Version"] = latest_standards_version()
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(
        wt, debian_path, source, compat_release=compat_release,
        buildsystem="octave", addons=['octave'])
    # For now, just assume a single binary package that is
    # architecture-independent.
    control.add_binary({
        "Package": "octave-%s" % metadata['Name'].lower(),
        "Architecture": 'all',
        'Depends': '${octave:Depends}, ${misc:Depends}',
        'Description': '${octave:Upstream-Description}',
        })
    return control


def process_default(es, session, wt, subpath, debian_path, upstream_version,
                    metadata, compat_release, buildsystem, buildsystem_subpath,
                    kickstart_from_dist):
    kickstart_from_dist(wt, subpath)
    control = es.enter_context(ControlEditor.create(
        wt.abspath(os.path.join(debian_path, 'control'))))
    source = control.source
    upstream_name = metadata['Name']
    source_name = source_name_from_upstream_name(upstream_name)
    if source_name is None:
        raise SourceNameUnknown(upstream_name)
    source["Source"] = source_name
    source["Rules-Requires-Root"] = "no"
    source["Standards-Version"] = latest_standards_version()
    build_deps, test_deps = get_project_wide_deps(
        session, wt, subpath, buildsystem, buildsystem_subpath)
    import_build_deps(source, build_deps)
    setup_debhelper(
        wt, debian_path,
        source, compat_release=compat_release)
    # For now, just assume a single binary package that is
    # architecture-dependent.
    for binary_name, arch in [(source['Source'], "any")]:
        control.add_binary({"Package": binary_name, "Architecture": arch})
    return control


def process_cargo(es, session, wt, subpath, debian_path, upstream_version,
                  metadata, compat_release, buildsystem, buildsystem_subpath,
                  kickstart_from_dist):
    wt.branch.generate_revision_history(NULL_REVISION)
    reset_tree(wt, wt.basis_tree(), subpath)
    from debmutate.debcargo import (
        DebcargoControlShimEditor,
        unmangle_debcargo_version,
    )
    crate = metadata.get('Cargo-Crate')
    if crate is None:
        crate = metadata['Name'].replace('_', '-')
    if not wt.has_filename(debian_path):
        wt.mkdir(debian_path)
    if not wt.is_versioned(debian_path):
        wt.add(debian_path)
    # Only set semver_suffix if this is not the latest version
    import semver
    try:
        desired_version = semver.VersionInfo.parse(upstream_version)
    except ValueError as exc:
        raise BuildSystemProcessError(buildsystem, str(exc), exc) from exc
    data = load_crate_info(crate)
    if data is None:
        raise BuildSystemProcessError(
            buildsystem, 'Crate does not exist' % crate)
    features = None
    crate_version = None
    for version_info in data['versions']:
        available_version = semver.VersionInfo.parse(version_info['num'])
        if ((available_version.major, available_version.minor)
                > (desired_version.major, desired_version.minor)):
            semver_suffix = True
            break
        if unmangle_debcargo_version(upstream_version) == version_info['num']:
            crate_version = version_info['num']
            features = list(version_info['features'])
    else:
        semver_suffix = False
    control = es.enter_context(DebcargoControlShimEditor.from_debian_dir(
        wt.abspath(debian_path), crate, crate_version, features))
    control.debcargo_editor['semver_suffix'] = semver_suffix
    control.debcargo_editor['overlay'] = '.'
    return control


Processor = Callable[
    [contextlib.ExitStack, Session, WorkingTree, str, str, str,
     Dict[str, Any], Optional[str], BuildSystem, str, Kickstarter],
    ControlEditor]


PROCESSORS: Dict[str, Processor] = {
    "setup.py": process_setup_py,
    "npm": process_npm,
    "maven": process_maven,
    "dist-zilla": process_dist_zilla,
    "dist-inkt": process_dist_zilla,
    "perl-build-tiny": process_perl_build_tiny,
    "makefile.pl": process_makefile_pl,
    "cargo": process_cargo,
    "golang": process_golang,
    "R": process_r,
    "octave": process_octave,
}


def source_name_from_directory_name(path):
    d = os.path.basename(path)
    if '-' in d:
        parts = d.split('-')
        if parts[-1][0].isdigit():
            return '-'.join(parts[:-1])
    return d


@dataclass
class DebianizeResult:
    """Debianize result."""

    upstream_branch_name: Optional[str] = None
    tag_names: Dict[Optional[str], RevisionID] = field(default_factory=dict)
    upstream_version: Optional[str] = None
    wnpp_bugs: List[Tuple[int, str]] = field(default_factory=list)
    vcs_url: Optional[str] = None


def import_build_deps(source, build_deps):
    for build_dep in build_deps:
        for rel in build_dep.relations:
            source["Build-Depends"] = ensure_relation(
                source.get("Build-Depends", ""),
                PkgRelation.str([rel]))


def import_upstream_dist(
        pristine_tar_source: BasePristineTarSource, wt: WorkingTree,
        upstream_source: UpstreamSource, subpath: str, source_name: str,
        upstream_version: str, session: Session
        ) -> Tuple[Tuple[RevisionID, str], Optional[str],
                   Dict[Optional[str], RevisionID]]:
    if pristine_tar_source.has_version(
            source_name, upstream_version, try_hard=False):
        logging.warning(
            'Upstream version %s/%s already imported.',
            source_name, upstream_version)
        pristine_revids = pristine_tar_source\
            .version_as_revisions(source_name, upstream_version)
        upstream_branch_name = None
        tag_names: Dict[Optional[str], RevisionID] = {}
    else:
        (pristine_revids, tag_names,
         upstream_branch_name) = import_upstream_version_from_dist(
            wt, subpath,
            upstream_source,
            source_name, upstream_version,
            session=session)

    assert (isinstance(pristine_revids[None], tuple)
            and len(pristine_revids[None])), repr(pristine_revids[None])
    return pristine_revids[None], upstream_branch_name, tag_names


def generic_get_source_name(wt, subpath, metadata):
    try:
        source_name = source_name_from_upstream_name(metadata['Name'])
    except KeyError:
        source_name = None
    else:
        if not valid_debian_package_name(source_name):
            source_name = None
    if source_name is None:
        source_name = source_name_from_directory_name(wt.abspath(subpath))
        if not valid_debian_package_name(source_name):
            source_name = None
    return source_name


def get_upstream_version(
        upstream_source, metadata: Dict[str, Any],
        local_dir=None,
        upstream_subpath: Optional[str] = None,
        upstream_version: Optional[str] = None) -> Tuple[str, str]:
    # TODO(jelmer): if upstream_subpath != "", perhaps ignore info from
    # upstream_source?
    if upstream_version is None:
        upstream_version, mangled_upstream_version = (
            upstream_source.get_latest_version(metadata.get("Name"), None))
    else:
        mangled_upstream_version = debianize_upstream_version(upstream_version)
    if upstream_version is None:
        raise NoUpstreamReleases(upstream_source, metadata.get("Name"))

    upstream_revision = upstream_source.version_as_revision(
        metadata.get("Name"), mangled_upstream_version)

    if upstream_version is None and "Version" in metadata:
        # They haven't done any releases yet. Assume we're ahead of
        # the next announced release?
        next_upstream_version = debian_upstream_version(metadata["Version"])
        upstream_version = upstream_version_add_revision(
            upstream_source.upstream_branch, next_upstream_version,
            upstream_revision, "~"
        )
    if upstream_version is None:
        upstream_version = upstream_version_add_revision(
            upstream_source.upstream_branch, "0", upstream_revision, "+"
        )
        logging.warning(
            "Unable to determine upstream version, using %s.",
            upstream_version)
    return upstream_version, mangled_upstream_version


def find_wnpp_bugs_harder(source_name, upstream_name):
    wnpp_bugs = find_wnpp_bugs(source_name)
    if not wnpp_bugs and source_name != upstream_name:
        wnpp_bugs = find_wnpp_bugs(upstream_name)
    if not wnpp_bugs:
        wnpp_bugs = find_archived_wnpp_bugs(source_name)
        if wnpp_bugs:
            logging.warning(
                "Found archived ITP/RFP bugs for %s: %r",
                source_name, [bug for bug, kind in wnpp_bugs]
            )
        else:
            logging.warning(
                "No relevant WNPP bugs found for %s", source_name)
    else:
        logging.info(
            "Found WNPP bugs for %s: %r",
            source_name, [bug for bug, kind in wnpp_bugs])

    return wnpp_bugs


def debianize(  # noqa: C901
    wt: WorkingTree, subpath: str,
    *,
    upstream_branch: Optional[Branch], upstream_subpath: Optional[str],
    use_inotify: Optional[bool] = None,
    diligence: int = 0,
    trust: bool = False,
    check: bool = False,
    net_access: bool = True,
    force_subprocess: bool = False,
    force_new_directory: bool = False,
    compat_release: Optional[str] = None,
    minimum_certainty: str = MINIMUM_CERTAINTY,
    consult_external_directory: bool = True,
    verbose: bool = False,
    schroot: Optional[str] = None,
    unshare: Optional[str] = None,
    create_dist=None,
    committer: Optional[str] = None,
    upstream_version_kind: str = "auto",
    debian_revision: str = "1",
    upstream_version: Optional[str] = None,
    requirement: Optional[Requirement] = None,
    team: Optional[str] = None,
    buildsystem_name: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None
):
    if committer is None:
        committer = get_committer(wt)

    debian_path = osutils.pathjoin(subpath, "debian")
    if (wt.has_filename(debian_path)
            and list(os.listdir(wt.abspath(debian_path)))
            and not force_new_directory):
        raise DebianDirectoryExists(wt.abspath(subpath))

    metadata_items: List[UpstreamDatum] = []
    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    def import_metadata_from_path(p):
        metadata_items.extend(guess_upstream_info(p, trust_package=trust))
        assert isinstance(metadata, dict)
        metadata.update(
            summarize_upstream_metadata(
                metadata_items, p, net_access=net_access,
                consult_external_directory=consult_external_directory,
                check=check))

    if not verbose:
        commit_reporter = NullCommitReporter()
    else:
        commit_reporter = None

    result = DebianizeResult()

    with wt.lock_write():
        with contextlib.ExitStack() as es:
            es.enter_context(ResetOnFailure(wt, subpath=subpath))

            if not wt.has_filename(debian_path):
                wt.mkdir(debian_path)
            if not wt.is_versioned(debian_path):
                wt.add(debian_path)

            session: Session
            if schroot:
                logging.info('Using schroot %s', schroot)
                session = SchrootSession(schroot)
            elif unshare:
                logging.info('Using tarball %s for unshare', unshare)
                from ognibuild.session.unshare import (
                    UnshareSession,  # type: ignore
                )
                session = UnshareSession.from_tarball(unshare)
            else:
                session = PlainSession()

            if upstream_branch:
                upstream_source = UpstreamBranchSource.from_branch(
                    upstream_branch, version_kind=upstream_version_kind,
                    local_dir=wt.controldir,
                    create_dist=(
                        create_dist or partial(default_create_dist, session)))
            else:
                upstream_source = None

            if upstream_version is not None:
                mangled_upstream_version = debianize_upstream_version(
                    upstream_version)
            else:
                upstream_version, mangled_upstream_version = (
                    get_upstream_version(
                        upstream_source, metadata, local_dir=wt.controldir,
                        upstream_subpath=upstream_subpath,
                        upstream_version=upstream_version))

            result.upstream_version = upstream_version

            source_name = generic_get_source_name(wt, subpath, metadata)

            def kickstart_from_dist(wt, subpath):
                logging.info(
                    "Kickstarting from dist tarball. "
                    "Using upstream version %s", mangled_upstream_version)

                pristine_tar_source = get_pristine_tar_source(wt, wt.branch)
                ((upstream_dist_revid, upstream_dist_subpath),
                 result.upstream_branch_name,
                 result.tag_names) = import_upstream_dist(
                    pristine_tar_source, wt, upstream_source, subpath,
                    source_name, mangled_upstream_version, session)

                if wt.branch.last_revision() != upstream_dist_revid:
                    wt.pull(
                        upstream_source.upstream_branch, overwrite=True,
                        stop_revision=upstream_dist_revid)

                    # Gather metadata items again now that we're at the correct
                    # revision
                    import_metadata_from_path(
                        wt.abspath(upstream_dist_subpath))

                if wt.has_filename(debian_path) and force_new_directory:
                    shutil.rmtree(wt.abspath(debian_path))
                    wt.mkdir(wt.abspath(debian_path))
                    with contextlib.suppress(PointlessCommit):
                        wt.commit(
                            'Remove old debian directory',
                            specific_files=[debian_path],
                            reporter=NullCommitReporter())

                wt.mkdir(os.path.join(debian_path, 'source'))
                wt.add(os.path.join(debian_path, 'source'))
                wt.put_file_bytes_non_atomic(
                    os.path.join(debian_path, 'source', 'format'),
                    b'3.0 (quilt)\n')

            if upstream_source:
                try:
                    (upstream_vcs_tree,
                     upstream_vcs_subpath) = upstream_source.revision_tree(
                        source_name, mangled_upstream_version)
                except (PackageVersionNotPresent, NoSuchRevision):
                    logging.warning(
                        'Unable to find upstream version %s/%s '
                        'in upstream source %r. Unable to extract metadata.',
                        source_name, mangled_upstream_version, upstream_source)
                    exported_upstream_tree_path = None
                else:
                    assert upstream_vcs_subpath == ''
                    # TODO(jelmer): Don't export, just access from memory.
                    exported_upstream_tree_path = es.enter_context(
                        TemporaryDirectory())
                    assert isinstance(upstream_subpath, str)
                    dupe_vcs_tree(
                        upstream_vcs_tree, exported_upstream_tree_path)
                    exported_upstream_tree_subpath = os.path.join(
                        exported_upstream_tree_path, upstream_subpath)
                    if not os.path.isdir(exported_upstream_tree_subpath):
                        raise Exception(
                            f'subdirectory {upstream_subpath} does not '
                            f'exist in upstream version {upstream_version}')
                    import_metadata_from_path(exported_upstream_tree_subpath)

            if (buildsystem_name is None
                    and exported_upstream_tree_path is not None):
                buildsystem_subpath, buildsystem = get_buildsystem(
                    os.path.join(exported_upstream_tree_path, subpath))
                if buildsystem:
                    buildsystem_name = buildsystem.name
            else:
                buildsystem = None
                buildsystem_subpath = ''

            if buildsystem_name:
                try:
                    process = PROCESSORS[buildsystem_name]
                except KeyError:
                    logging.warning(
                        'No support in debianize for build system %s, '
                        'falling back to default.',
                        buildsystem_name)
                    process = process_default
            else:
                process = process_default

            logging.info('Creating core packaging using %s', process.__name__)

            os.chdir(wt.abspath(subpath))

            control = process(
                es, session,
                wt, subpath,
                debian_path,
                upstream_version,
                metadata, compat_release,
                buildsystem,
                buildsystem_subpath,
                kickstart_from_dist)

            source = control.source

            if team:
                control.source['Maintainer'] = team
                uploader = '{} <{}>'.format(*get_maintainer())
                if uploader != team:
                    control.source['Uploaders'] = uploader

            if not valid_debian_package_name(source['Source']):
                raise SourcePackageNameInvalid(source['Source'])

            if net_access:
                wnpp_bugs = find_wnpp_bugs_harder(
                    source['Source'], metadata.get('Name'))
            else:
                wnpp_bugs = None

            result.wnpp_bugs = wnpp_bugs

            version = Version(mangled_upstream_version + "-" + debian_revision)
            write_changelog_template(
                wt.abspath(os.path.join(debian_path, "changelog")),
                source["Source"],
                version,
                get_maintainer(),
                wnpp_bugs,
            )

            if (requirement and requirement.family == 'apt' and
                    not cast(AptRequirement, requirement).satisfied_by(
                        control.binaries, version)):
                logging.warning(
                    'Debianized package (binary packages: %r), version %s '
                    'did not satisfy requirement %r. ',
                    [binary['Package'] for binary in control.binaries],
                    version, requirement)
                if upstream_branch:
                    logging.info('Wrong repository (%s)?', upstream_branch)
                raise DebianizedPackageRequirementMismatch(
                    requirement, control, version,
                    upstream_branch)

            control.wrap_and_sort()
            control.sort_binary_packages()

        wt.smart_add([wt.abspath(debian_path)])
        wt.commit(
            "Create debian/ directory",
            allow_pointless=False,
            committer=committer,
            reporter=commit_reporter,
        )

    with wt.lock_write():

        lintian_fixers = available_lintian_fixers(
            force_subprocess=force_subprocess)

        run_lintian_fixers(
            wt,
            list(lintian_fixers),
            update_changelog=False,
            compat_release=compat_release,
            verbose=verbose,
            minimum_certainty=minimum_certainty,
            trust_package=trust,
            allow_reformatting=True,
            use_inotify=use_inotify,
            subpath=subpath,
            net_access=net_access,
            opinionated=True,
            diligence=diligence,
        )

        try:
            result.vcs_url = unsplit_vcs_url(*update_official_vcs(
                wt, subpath=subpath, committer=committer))
        except NoVcsLocation:
            logging.debug(
                'No public VCS location specified and unable to guess it '
                'based on maintainer e-mail.')
        except FileNotFoundError:
            logging.info(
                'No control file or debcargo.toml file, '
                'not setting vcs information.')

    return result


class SimpleTrustedAptRepo:

    def __init__(self, directory):
        self.directory = directory
        self.httpd = None
        self.thread = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False

    def sources_lines(self):
        if not self.httpd:
            raise RuntimeError("httpd not yet started")
        if os.path.exists(os.path.join(self.directory, 'Packages.gz')):
            return [
                "deb [trusted=yes] http://%s:%d/ ./" % (
                    self.httpd.server_name, self.httpd.server_port)]
        else:
            return []

    def start(self):
        if self.thread is not None:
            raise RuntimeError('thread already active')
        import http.server
        from threading import Thread
        hostname = "localhost"

        class RequestHandler(http.server.SimpleHTTPRequestHandler):

            def log_message(self, format, *args):
                return

        handler = partial(RequestHandler, directory=self.directory)
        self.httpd = http.server.HTTPServer((hostname, 0), handler, False)
        self.httpd.server_bind()
        logging.info(
            'Local apt repo started at http://%s:%d/',
            self.httpd.server_name, self.httpd.server_port)
        self.httpd.server_activate()

        def serve_forever(httpd):
            with httpd:  # to make sure httpd.server_close is called
                httpd.serve_forever()

        self.thread = Thread(
            target=serve_forever, args=(self.httpd, ), daemon=True)
        self.thread.start()

    def stop(self):
        if self.httpd is not None:
            self.httpd.shutdown()
            self.httpd = None
        if self.thread is not None:
            self.thread.join()
            self.thread = None

    def refresh(self):
        import gzip
        packages = subprocess.check_output(
            ['dpkg-scanpackages', '-m', '.', '/dev/null'],
            cwd=self.directory)
        with gzip.GzipFile(
                os.path.join(self.directory, 'Packages.gz'), 'wb') as f:
            f.write(packages)


def use_packaging_branch(wt: WorkingTree, branch_name: str) -> None:
    last_revision = wt.last_revision()
    try:
        target_branch = wt.controldir.open_branch(branch_name)
    except NotBranchError:
        target_branch = wt.controldir.create_branch(branch_name)

    target_branch.generate_revision_history(last_revision)
    logging.info('Switching to packaging branch %s.', branch_name)
    wt.controldir.set_branch_reference(target_branch, name="")
    # TODO(jelmer): breezy bug?
    wt._branch = target_branch


class DebianizeFixer(BuildFixer):
    """Fixer that invokes debianize to create a package."""

    def __str__(self):
        return "debianize fixer"

    def __repr__(self):
        return "{}({!r}, {!r})".format(
            type(self).__name__, self.vcs_directory,
            self.apt_repo)

    def __init__(self, vcs_directory, apt_repo, do_build, diligence=0,
                 trust=False, check=True, net_access=True,
                 force_new_directory=False,
                 team=None, verbose=False, force_subprocess=False,
                 upstream_version_kind="auto",
                 debian_revision="1",
                 schroot=None,
                 unshare=None,
                 consult_external_directory=True,
                 use_inotify=None,
                 create_dist=None,
                 compat_release=None):
        self.vcs_directory = vcs_directory
        self.apt_repo = apt_repo
        self.diligence = diligence
        self.trust = trust
        self.check = check
        self.net_access = net_access
        self.force_new_directory = force_new_directory
        self.team = team
        self.verbose = verbose
        self.force_subprocess = force_subprocess
        self.upstream_version_kind = upstream_version_kind
        self.debian_revision = debian_revision
        self.consult_external_directory = consult_external_directory
        self.schroot = schroot
        self.unshare = unshare
        self.use_inotify = use_inotify
        self.create_dist = create_dist
        self.compat_release = compat_release
        self.do_build = do_build

    def can_fix(self, problem):
        requirement = problem_to_upstream_requirement(problem)
        if requirement is None:
            return False
        return find_upstream(requirement) is not None

    def _fix(self, problem, phase: Tuple[str, ...]):
        requirement = problem_to_upstream_requirement(problem)
        if requirement is None:
            return False
        logging.debug(
            'Translated problem %r to requirement %r', problem, requirement)
        upstream_info = find_upstream(requirement)
        if upstream_info is None:
            logging.error(
                'Unable to find upstream information for requirement %r',
                requirement)
            return False
        if upstream_info.branch_url:
            logging.info(
                'Packaging %r to address %r',
                upstream_info.branch_url, problem)
            try:
                upstream_branch = Branch.open(upstream_info.branch_url)
            except NotBranchError as e:
                logging.warning(
                    'Unable to open branch %s: %r', upstream_info.branch_url,
                    e)
                upstream_branch = None
        else:
            upstream_branch = None
        if upstream_info.name is not None:
            vcs_path = os.path.join(
                self.vcs_directory,
                upstream_info.name.replace('/', '-'))
        else:
            raise AssertionError('no upstream name provided')
        if os.path.exists(vcs_path):
            shutil.rmtree(vcs_path)
        if upstream_branch:
            format = upstream_branch.controldir.cloning_metadir()
        else:
            # TODO(jelmer): default to git?
            format = None
        result = ControlDir.create_branch_convenience(
            vcs_path, force_new_tree=True,
            format=format)
        new_wt = result.controldir.open_workingtree()
        new_subpath = ''
        debianize(
            new_wt, new_subpath,
            upstream_branch=upstream_branch,
            upstream_subpath=(upstream_info.branch_subpath or ''),
            use_inotify=self.use_inotify,
            diligence=self.diligence,
            create_dist=self.create_dist,
            trust=self.trust,
            check=self.check,
            net_access=self.net_access,
            force_new_directory=self.force_new_directory,
            force_subprocess=self.force_subprocess,
            compat_release=self.compat_release,
            consult_external_directory=self.consult_external_directory,
            verbose=self.verbose, schroot=self.schroot,
            unshare=self.unshare,
            debian_revision=self.debian_revision,
            upstream_version=upstream_info.version,
            upstream_version_kind=self.upstream_version_kind,
            requirement=requirement,
            buildsystem_name=upstream_info.buildsystem,
            team=self.team,
            metadata=upstream_info.metadata)
        self.do_build(
            new_wt, new_subpath, self.apt_repo.directory,
            extra_repositories=self.apt_repo.sources_lines())
        self.apt_repo.refresh()
        return True


def report_fatal(code, description, hint=None, details=None):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'versions': versions_dict(),
                'description': description,
                'details': details}, f)
    logging.fatal('%s', description)
    if hint:
        logging.info('%s', hint)


default_debianize_cache_dir = _lintian_brush_rs.default_debianize_cache_dir


def main(*, verbose: bool, directory: str, compat_release: str | None,
         dist_command: str, upstream: str | None, debian_binary: bool,
         debian_branch: str, disable_inotify: bool, diligence: int,
         trust: bool, check: bool, disable_net_access: bool,
         force_new_directory: bool, force_subprocess: bool,
         consult_external_directory: bool,
         schroot: str | None, unshare: str, upstream_version_kind: str,
         debian_revision: str, upstream_version: str, install: bool,
         iterate_fix: bool, output_directory: str, discard_output: bool,
         build_command: str, dep_server_url: str | None,
         max_build_iterations: int, recursive: bool,
         team: str | None):  # noqa: C901
    import breezy
    breezy.initialize()  # type: ignore
    import breezy.bzr  # noqa: E402
    import breezy.git  # noqa: E402

    if verbose:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO
    logging.basicConfig(level=loglevel, format='%(message)s')

    # pysimplesoap is quite verbose
    logging.getLogger('pysimplesoap.client').setLevel(logging.ERROR)

    if compat_release is None:
        import distro_info

        debian_info = distro_info.DebianDistroInfo()
        compat_release = debian_info.stable()

    try:
        wt, subpath = WorkingTree.open_containing(directory)
    except NotBranchError as e:
        logging.fatal(
            'please run debianize in an existing branch where '
            'it should add the packaging: %s', e)
        return 1

    create_dist_fn: Optional[Callable[[Tree, str, str, str], str]]
    if dist_command:
        def create_dist_fn(tree, package, version, target_dir, subpath=""):
            return run_dist_command(
                tree, package, version, target_dir, dist_command,
                subpath=subpath)
    else:
        create_dist_fn = None

    metadata = {}

    # For now...
    if upstream:
        try:
            upstream_branch, upstream_subpath = Branch.open_containing(
                upstream)
        except NotBranchError as e:
            logging.fatal('%s: not a valid branch: %s', upstream, e)
            return 1
        metadata['Repository'] = UpstreamDatum(
            'Repository', upstream, certainty='confident')
    elif debian_binary:
        apt_requirement = AptRequirement.from_str(debian_binary)
        upstream_info = find_apt_upstream(apt_requirement)
        if not upstream_info:
            logging.fatal(
                '%s: Unable to find upstream info for %s', debian_binary,
                apt_requirement)
            return 1
        logging.info(
            'Found relevant upstream branch at %s', upstream_info.branch_url)
        upstream_branch = Branch.open(upstream_info.branch_url)
        upstream_subpath = upstream_info.branch_subpath

        if upstream_info.branch_url:
            metadata['Repository'] = UpstreamDatum(
                'Repository', upstream_info.branch_url, certainty='confident')
    else:
        if wt.has_filename(os.path.join(subpath, 'debian')):
            report_fatal(
                code='debian-directory-exists',
                description=(
                    f"{wt.abspath(subpath)}: "
                    "A debian directory already exists."),
                hint=("Run lintian-brush instead or "
                      "specify --force-new-directory."),
            )
            return 1
        logging.info(
            'No upstream repository specified, using upstream source in %s',
            wt.abspath(subpath))
        upstream_branch = wt.branch
        upstream_subpath = subpath

    if debian_branch:
        from debmutate.vendor import get_vendor_name

        use_packaging_branch(
            wt, debian_branch % {'vendor': get_vendor_name().lower()})

    use_inotify = (False if disable_inotify else None)
    with wt.lock_write():
        try:
            debianize_result = debianize(
                wt, subpath,
                upstream_branch=upstream_branch,
                upstream_subpath=upstream_subpath,
                use_inotify=use_inotify,
                diligence=diligence,
                trust=trust,
                check=check,
                net_access=not disable_net_access,
                force_new_directory=force_new_directory,
                force_subprocess=force_subprocess,
                compat_release=compat_release,
                consult_external_directory=consult_external_directory,
                verbose=verbose,
                schroot=schroot,
                unshare=unshare,
                upstream_version_kind=upstream_version_kind,
                debian_revision=debian_revision,
                create_dist=create_dist_fn,
                upstream_version=upstream_version,
                metadata=metadata,
            )
        except PackageVersionNotPresent:
            if upstream_version:
                report_fatal(
                    'requested-version-missing',
                    'Requested version %s not present upstream' %
                    upstream_version)
                return 1
            else:
                # For now
                raise
        except DistCommandFailed as e:
            report_fatal(
                e.kind or "dist-command-failed", e.error)  # type: ignore
            return 1
        except WorkspaceDirty:
            report_fatal(
                "pending-changes", "Please commit pending changes first.")
            return 1
        except DebianDirectoryExists as e:
            report_fatal(
                code='debian-directory-exists',
                description="%s: A debian directory already exists. " % e.path,
                hint=("Run lintian-brush instead or "
                      "specify --force-new-directory."),
            )
            return 1
        except SourcePackageNameInvalid as e:
            report_fatal(
                code='invalid-source-package-name',
                description=(
                    "Generated source package name %r is not valid."
                    % e.source))
            return 1
        except NoUpstreamReleases:
            report_fatal(
                'no-upstream-releases',
                'The upstream project does not appear to have '
                'made any releases.')
        except NoBuildToolsFound:
            report_fatal(
                'no-build-tools',
                "Unable to find any build systems in upstream sources.")
            return 1
        except DetailedFailure as e:
            report_fatal(
                'debianize-' + e.error.kind,
                str(e),
                details=(e.error.json() if e.error else None))
            return 1
        except BuildSystemProcessError as e:
            report_fatal(
                'build-system-process-error',
                e.message)
            return 1
        except OSError as e:
            if e.errno == errno.ENOSPC:
                report_fatal('no-space-on-device', str(e))
                return 1
            else:
                raise

    if install:
        iterate_fix = True

    if iterate_fix:
        session: Session

        if schroot:
            logging.info('Using schroot %s', schroot)
            session = SchrootSession(schroot)
        elif unshare:
            logging.info('Using tarball %s for unshare', unshare)
            from ognibuild.session.unshare import (
                UnshareSession,  # type: ignore
            )
            session = UnshareSession.from_tarball(unshare)
        else:
            session = PlainSession()

        with contextlib.ExitStack() as es:
            es.enter_context(session)
            apt = AptManager.from_session(session)
            if discard_output:
                output_directory = es.enter_context(TemporaryDirectory())
            if not output_directory:
                output_directory = default_debianize_cache_dir()
                logging.info(
                    'Building dependencies in %s', output_directory)

            def do_build(
                    wt, subpath, incoming_directory, extra_repositories=None):
                fixers = default_fixers(
                    wt, subpath, apt,
                    update_changelog=False, committer=None,
                    dep_server_url=dep_server_url)
                return build_incrementally(
                    local_tree=wt,
                    suffix=None,
                    build_suite=None,
                    fixers=fixers,
                    output_directory=incoming_directory,
                    build_command=build_command,
                    build_changelog_entry=None,
                    max_iterations=max_build_iterations,
                    subpath=subpath,
                    extra_repositories=extra_repositories,
                )

            try:
                if recursive:
                    vcs_directory = os.path.join(output_directory, 'vcs')
                    os.makedirs(vcs_directory, exist_ok=True)
                    apt_directory = os.path.join(output_directory, 'apt')
                    os.makedirs(apt_directory, exist_ok=True)

                    def main_build():
                        return do_build(
                            wt, subpath, apt_repo.directory,
                            extra_repositories=apt_repo.sources_lines())
                    with SimpleTrustedAptRepo(apt_directory) as apt_repo:
                        (changes_names, _cl_entry) = iterate_with_build_fixers(
                            [DebianizeFixer(
                                vcs_directory, apt_repo,
                                diligence=diligence,
                                trust=trust,
                                check=check,
                                net_access=not disable_net_access,
                                force_new_directory=force_new_directory,
                                team=team, verbose=verbose,
                                force_subprocess=force_subprocess,
                                upstream_version_kind=(
                                    upstream_version_kind),
                                debian_revision=debian_revision,
                                schroot=schroot,
                                unshare=unshare, use_inotify=use_inotify,
                                consult_external_directory=(
                                    consult_external_directory),
                                create_dist=create_dist_fn,
                                compat_release=compat_release,
                                do_build=do_build)],
                            main_build)
                else:
                    (changes_names, _cl_entry) = do_build(
                        wt, subpath, output_directory)
            except DetailedDebianBuildFailure as e:
                if e.phase is None:
                    phase = 'unknown phase'
                elif len(e.phase) == 1:
                    phase = e.phase[0]
                else:
                    phase = f'{e.phase[0]} ({e.phase[1]})'
                logging.fatal('Error during %s: %s', phase, e.error)
                return 1
            except UnidentifiedDebianBuildError as e:
                if e.phase is None:
                    phase = 'unknown phase'
                elif len(e.phase) == 1:
                    phase = e.phase[0]
                else:
                    phase = f'{e.phase[0]} ({e.phase[1]})'
                logging.fatal('Error during %s: %s', phase, e.description)
                return 1
            except DebianizedPackageRequirementMismatch as e:
                if upstream_branch:
                    hint = 'Wrong repository (%s)?' % e.upstream_branch
                else:
                    hint = None
                report_fatal(
                    'package-requirements-mismatch',
                    'Debianized package (binary packages: {!r}), version {} '
                    'did not satisfy requirement {!r}. '.format(
                        [binary['Package'] for binary in e.control.binaries],
                        e.version, e.requirement), hint=hint)
                return 1
            logging.info('Built %r.', changes_names)
            if install:
                subprocess.check_call(
                    ["debi"] + [
                        os.path.join(output_directory, cn)
                        for cn in changes_names])

    if debianize_result.vcs_url:
        target_branch_url = vcs_git_url_to_bzr_url(
            debianize_result.vcs_url)
    else:
        target_branch_url = None

    if svp_enabled():
        with open(os.environ['SVP_RESULT'], "w") as f:
            json.dump({
                "description": "Debianized package",
                "target-branch-url": target_branch_url,
                "versions": versions_dict(),
                "context": {
                    "wnpp_bugs": debianize_result.wnpp_bugs,
                }
            }, f)

    return 0



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError
from upstream_ontologist import (
    UpstreamDatum,
)
from upstream_ontologist.guess import (
    guess_upstream_metadata_items,
)

from lintian_brush import min_certainty
from lintian_brush.fixer import (
    meets_minimum_certainty,
    report_result,
    trust_package,
)

fields = []
achieved_certainty = []


def add_upstream_metadata(copyright):
    if copyright.header.upstream_name and copyright.header.upstream_contact:
        return
    import ruamel.yaml
    try:
        with open('debian/upstream/metadata') as f:
            inp = f.read()
    except FileNotFoundError:
        upstream_metadata = {}
    else:
        code = ruamel.yaml.safe_load(inp)
        upstream_metadata = {
            k: UpstreamDatum(k, v, 'certain') for (k, v) in code.items()}
    for datum in guess_upstream_metadata_items(
            '.', trust_package=trust_package()):
        if not meets_minimum_certainty(datum.certainty):
            continue
        if (datum.field not in upstream_metadata or
                upstream_metadata[datum.field].certainty != 'certain'):
            upstream_metadata[datum.field] = datum

    if not copyright.header.upstream_name:
        try:
            datum = upstream_metadata['Name']
        except KeyError:
            pass
        else:
            if datum.value:
                copyright.header.upstream_name = datum.value
                fields.append('Upstream-Name')
                achieved_certainty.append(datum.certainty)

    if not copyright.header.upstream_contact:
        try:
            datum = upstream_metadata['Contact']
        except KeyError:
            pass
        else:
            if datum.value:
                value = datum.value
                if isinstance(value, str):
                    value = [value]
                copyright.header.upstream_contact = value
                fields.append('Upstream-Contact')
                achieved_certainty.append(datum.certainty)


with suppress(FileNotFoundError, NotMachineReadableError), \
        CopyrightEditor() as updater:
    add_upstream_metadata(updater.copyright)

certainty = min_certainty(achieved_certainty)

if len(fields) == 1:
    report_result('Set field %s in debian/copyright.' % ', '.join(fields),
                  certainty=certainty)
else:
    report_result('Set fields %s in debian/copyright.' % ', '.join(fields),
                  certainty=certainty)



#!/usr/bin/python3

from email.utils import parseaddr

from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    if ("Maintainer" in updater.source and
            parseaddr(updater.source["Maintainer"])[1] ==
            'packages@qa.debian.org' and 'Uploaders' in updater.source):
        issue = LintianIssue(
            updater.source, 'uploaders-in-orphan', info=())
        if issue.should_fix():
            del updater.source["Uploaders"]
            issue.report_fixed()


report_result("Remove uploaders from orphaned package.")



#!/usr/bin/python3
# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import logging
import re
import sys
import warnings
from dataclasses import dataclass
from typing import Dict, Set

import yaml
from debmutate.control import suppress_substvar_warnings

from debian.deb822 import PkgRelation


def parse_relations(t: str):
    with warnings.catch_warnings():
        suppress_substvar_warnings()
        return PkgRelation.parse_relations(t)


REGEXES = [
    r'.*\((.*, )?(dummy )?transitional (dummy )?package\)',
    r'.*\((.*, )?(dummy )?transitional (dummy )?package for ([^ ]+)\)',
    r'.*\(transitional development files\)',
    r'.*\(transitional\)',
    r'.* [-—] transitional( package)?',
    r'.*\[transitional package\]',
    r'.* - transitional (dummy )?package',
    r'transitional package -- safe to remove',
    r'(dummy )?transitional (dummy )?package (for|to) (.*)',
    r'transitional dummy package',
    r'transitional dummy package: ([^ ]+)',
    r'transitional package, ([^ ]+)',
    r'(dummy )?transitional (dummy )?package, ([^ ]+) to ([^ ]+)',
    r'transitional package( [^ ]+)?',
    r'([^ ]+) transitional package',
    r'.* transitional package',
    r'.*transitional package for .*',
]


@dataclass
class TransitionalPackage:

    from_name: str
    to_expr: str

    def json(self):
        return {'from_name': self.from_name, 'to_expr': self.to_expr}


def find_reverse_dependencies(udd, package):
    by_source: Dict[str, Set[str]] = {}
    fields = ['recommends', 'depends', 'pre_depends', 'enhances',
              'suggests', 'provides']
    query = (
        "SELECT source, package, " + ", ".join(fields) +
        " FROM packages WHERE " + " OR ".join([
            f"{field} LIKE CONCAT('%%', %s::text, '%%')" for field in fields]))
    with udd.cursor() as cursor:
        cursor.execute(query, (package, ))
        for row in cursor:
            for relname in fields:
                parsed = parse_relations(row[relname] or '')
                for option in parsed:
                    for rel in option:
                        if rel['name'] == package:
                            by_source.setdefault(
                                row['source'], set()).add(row['package'])

    fields = ["build_depends", "build_depends_indep", "build_depends_arch",
              "build_conflicts", "build_conflicts_indep",
              "build_conflicts_arch"]
    query = ("SELECT source, " + ", ".join(fields) +
             " FROM sources WHERE " + " OR ".join([
                 f"{field} LIKE CONCAT('%%', %s::text, '%%')"
                 for field in fields]))
    with udd.cursor() as cursor:
        cursor.execute(query, (package, ))
        for row in cursor:
            for relname in fields:
                parsed = parse_relations(row[relname] or '')
                for option in parsed:
                    for rel in option:
                        if rel['name'] == package:
                            by_source.setdefault(row['source'], set())
    return by_source


def find_dummy_transitional_packages(udd, release):
    ret = {}
    query = """\
SELECT package, description, depends FROM packages
WHERE release = %s AND description LIKE '%%transitional%%'"""
    with udd.cursor() as cursor:
        cursor.execute(query, (release, ))
        for row in cursor:
            for regex in REGEXES:
                if re.fullmatch(regex, row[1]):
                    break
            else:
                logging.debug(
                    'Unknown syntax for dummy package description: %r', row[1])
            if row[2] is None:
                logging.debug('no replacement for %s', row[0])
            else:
                depends = parse_relations(row[2])
                if len(depends) != 1:
                    logging.debug(
                        'no single transition target for %s: %r',
                        row[0], row[2])
            ret[row[0]] = TransitionalPackage(
                from_name=row[0], to_expr=row[2])  # type: ignore
        return ret


def main(argv=None):
    from .udd import connect_udd_mirror
    parser = argparse.ArgumentParser(prog="candidates")
    parser.add_argument("--release", type=str, default="sid")
    parser.add_argument('--list-transitional-dummy', action="store_true")
    parser.add_argument('--list-uses-transitional-dummy', action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    udd = connect_udd_mirror()
    transitions = find_dummy_transitional_packages(udd, args.release)

    if args.list_transitional_dummy:
        yaml.dump(
            [transition.json() for transition in transitions.values()],
            sys.stdout)
    elif args.list_uses_transitional_dummy:
        for dep in transitions:
            by_source = find_reverse_dependencies(udd, dep)
            for source, binaries in by_source.items():
                for binary in binaries:
                    print(f'{source} / {binary} / {dep}')
    else:
        parser.print_usage()
        return 1


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3

import sys

from debmutate.control import delete_from_list

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

# TODO(jelmer): support checking debcargo's maintainer/uploaders fields

try:
    with control as updater:
        if 'Uploaders' in updater.source:
            uploaders = updater.source['Uploaders'].split(',')
            maintainer = updater.source['Maintainer']
            if maintainer in [uploader.strip() for uploader in uploaders]:
                updater.source['Uploaders'] = delete_from_list(
                    updater.source['Uploaders'], maintainer)
                if not updater.source['Uploaders'].strip():
                    del updater.source['Uploaders']
                fixed_lintian_tag(
                    updater.source, 'maintainer-also-in-uploaders')
except FileNotFoundError:
    sys.exit(0)


report_result("Remove maintainer from uploaders.")



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for accessing the VCS Watch database."""


class VcsWatchError(Exception):
    """Error from vcswatch."""


class VcsWatch:
    """Read VcsWatch data through UDD."""

    def __init__(self):
        self._conn = None

    def __aenter__(self):
        from .udd import connect_udd_mirror
        self._conn = connect_udd_mirror()

    def __aexit__(self, exc_type, exc, tb):
        if self._conn is None:
            raise RuntimeError('not in context manager')
        return self._conn.__exit__(exc_type, exc, tb)

    def get_package(self, name):
        """Get the VCS information for a package.

        Args:
          name: Package name
        Returns:
          Tuple with (vcs_type, vcs_url, vcs_browser)
        """
        assert self._conn is not None, "call connect() first"
        with self._conn.cursor() as cursor:
            cursor.execute("""
select vcs, url, browser, status, error from vcswatch
where source = %s""", (name, ))
            row = cursor.fetchone()
            if row is None:
                raise KeyError(name)
            if row[3] == "ERROR":
                raise VcsWatchError(row[4])
            return row[:3]

    def get_branch_from_url(self, vcs, url):
        """Get the branch for a VCS URL."""
        assert self._conn is not None, "call connect() first"
        with self._conn.cursor() as cursor:
            cursor.execute(
                "select branch, status, error from vcswatch "
                "where url = %s and vcs = %s", (url, vcs))
            row = cursor.fetchone()
            if row is None:
                raise KeyError(url)
            if row[1] == "ERROR":
                raise VcsWatchError(row[2])
            return row[0]



#!/usr/bin/python3

import os
import re
import textwrap
from contextlib import suppress
from typing import Dict

from debmutate.copyright import CopyrightEditor

from debian.copyright import License, NotMachineReadableError
from lintian_brush.fixer import (
    fixed_lintian_tag,
    report_result,
    warn,
)
from lintian_brush.licenses import (
    COMMON_LICENSES_DIR,
    FULL_LICENSE_NAME,
)

# In reality, what debian ships as "/usr/share/common-licenses/BSD" is
# BSD-3-clause in SPDX.
SPDX_RENAMES = {
    'BSD': 'BSD-3-clause',
    }
CANONICAL_NAMES = {
    'CC0': 'CC0-1.0',
}
updated = set()
_common_licenses: Dict[str, str] = {}


def normalize_license_text(text):
    return re.sub('[\n\t ]+', ' ', text).strip()


def license_text_matches(text1, text2):
    return text1 == text2


def cached_common_license(name):
    try:
        return _common_licenses[name]
    except KeyError:
        with open(os.path.join(COMMON_LICENSES_DIR, name)) as f:
            _common_licenses[name] = normalize_license_text(
                f.read())
        return _common_licenses[name]


_COMMON_LICENSES = [
    ('CC0-1.0', cached_common_license('CC0-1.0').replace('Legal Code ', '')),
] + [(SPDX_RENAMES.get(name, name), cached_common_license(name))
     for name in os.listdir(COMMON_LICENSES_DIR)]


_BLURB = {
    'CC0-1.0': """\
To the extent possible under law, the author(s) have dedicated all copyright
and related and neighboring rights to this software to the public domain
worldwide. This software is distributed without any warranty.

You should have received a copy of the CC0 Public Domain Dedication along with
this software. If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.
""",
    'Apache-2.0': """\
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
""",
    'GPL-2+': """\
This package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>
""",
    'GPL-3+': """\
This package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>
""",
}


def find_common_license_from_fulltext(text):
    # Don't bother for anything that's short
    if len(text.splitlines()) < 15:
        return None
    text = normalize_license_text(text)
    text = drop_debian_file_reference(text) or text
    for shortname, fulltext in _COMMON_LICENSES:
        if license_text_matches(fulltext, text):
            return shortname


def drop_debian_file_reference(shorttext):
    try:
        i = shorttext.lower().index("on debian systems, ")
    except ValueError:
        return None
    return shorttext[:i].strip()


def debian_file_reference(name, filename):
    return '\n'.join(textwrap.wrap(f"""\
On Debian systems, the full text of the {name}
can be found in the file `/usr/share/common-licenses/{filename}'.
""", width=78))


def find_common_license_from_blurb(text):
    text = normalize_license_text(text)
    text_without_debian_reference = drop_debian_file_reference(text)
    for name, shorttext in _BLURB.items():
        if normalize_license_text(shorttext) in (
                text, text_without_debian_reference):
            return name


def canonical_license_id(license_id):
    # From the standard:
    #  For licenses that have multiple versions in use, the short name is
    #  formed from the general short name of the license family, followed by a
    #  dash and the version number. If the version number is omitted, the
    #  lowest version number is implied. When the license grant permits using
    #  the terms of any later version of that license, add a plus sign to the
    #  end of the short name. For example, the short name GPL refers to the GPL
    #  version 1 and is equivalent to GPL-1, although the latter is clearer and
    #  therefore preferred. If the package may be distributed under the GPL
    #  version 1 or any later version, use a short name of GPL-1+.
    #
    #  For SPDX compatibility, versions with trailing dot-zeroes are considered
    #  to be equivalent to versions without (e.g., “2.0.0” is considered equal
    #  to “2.0” and “2”).
    m = re.fullmatch(r'([A-Za-z0-9]+)(\-[0-9\.]+)?(\+)?', license_id)
    if not m:
        warn('Unable to get canonical name for %r' % license_id)
        return license_id
    version = (m.group(2) or '-1')[1:]
    while version.endswith('.0'):
        version = version[:-2]
    return '{}-{}{}'.format(m.group(1), version, m.group(3) or '')


renames = {}


def replace_full_license(para):
    license = para.license
    license_matched = find_common_license_from_fulltext(license.text)
    if license_matched is None:
        if os.path.exists(os.path.join(COMMON_LICENSES_DIR, license.synopsis)):
            warn(
                'A common license shortname (%s) is used, but license '
                'text not recognized.' % license.synopsis)
        return
    # The full license text was found. Replace it with a blurb.
    canonical_id = canonical_license_id(license.synopsis)
    for shortname, blurb in _BLURB.items():  # noqa: B007
        if canonical_id == canonical_license_id(shortname):
            break
    else:
        if license.synopsis in SPDX_RENAMES:
            renames[license.synopsis] = SPDX_RENAMES[license.synopsis]
            return
        else:
            warn(f'Found full license text for {license_matched}, '
                 f'but unknown synopsis {license.synopsis} ({canonical_id})')
        return
    if license_matched == 'Apache-2.0':
        fixed_lintian_tag(
            'source', 'copyright-file-contains-full-apache-2-license',
            info=())
    if license_matched.startswith('GFDL-'):
        fixed_lintian_tag(
            'source', 'copyright-file-contains-full-gfdl-license',
            info=())
    if license_matched.startswith('GPL-'):
        fixed_lintian_tag(
            'source', 'copyright-file-contains-full-gpl-license',
            info=())
    para.license = License(license.synopsis, blurb)
    return license_matched


def reference_common_license(para):
    license = para.license
    common_license = find_common_license_from_blurb(license.text)
    if not common_license:
        return
    if COMMON_LICENSES_DIR in license.text:
        return
    if para.comment is not None and COMMON_LICENSES_DIR in para.comment:
        return
    if 'X-Comment' in para and COMMON_LICENSES_DIR in para['X-Comment']:
        return
    if 'License-Reference' in para:
        return
    para.license = License(
        license.synopsis, license.text + '\n\n' + debian_file_reference(
            FULL_LICENSE_NAME.get(common_license, common_license),
            common_license))
    if common_license in ('Apache-2.0', 'Apache-2'):
        fixed_lintian_tag(
            'source', 'copyright-not-using-common-license-for-apache2',
            info=())
    elif common_license.startswith('GPL-'):
        fixed_lintian_tag(
            'source', 'copyright-not-using-common-license-for-gpl', info=())
    elif common_license.startswith('LGPL-'):
        fixed_lintian_tag(
            'source', 'copyright-not-using-common-license-for-lgpl', info=())
    elif common_license.startswith('GFDL-'):
        fixed_lintian_tag(
            'source', 'copyright-not-using-common-license-for-gfdl', info=())
    fixed_lintian_tag(
        'source', 'copyright-does-not-refer-to-common-license-file', info=())
    if license.synopsis != common_license:
        renames[license.synopsis] = common_license
    return common_license


with suppress(NotMachineReadableError, FileNotFoundError), \
        CopyrightEditor() as updater:
    for para in updater.copyright.all_paragraphs():
        license = para.license
        if not license or not license.text:
            continue
        replaced_license = replace_full_license(para)
        if replaced_license:
            updated.add(replaced_license)
        replaced_license = reference_common_license(para)
        if replaced_license:
            updated.add(replaced_license)
    for paragraph in updater.copyright.all_paragraphs():
        if not paragraph.license or not paragraph.license.synopsis:
            continue
        try:
            newsynopsis = renames[paragraph.license.synopsis]
        except KeyError:
            continue
        paragraph.license = License(newsynopsis, paragraph.license.text)


done = []
if updated:
    done.append(
        'refer to common license file for %s' % ', '.join(sorted(updated)))
if set(renames.values()) - set(updated):
    done.append('use common license names: ' + ', '.join(
        [f'{new} (was: {old})' for (old, new) in sorted(renames.items())
         if new not in updated]))


if done:
    report_result(
        done[0][0].capitalize() + ('; '.join(done) + '.')[1:])



#!/usr/bin/python3
from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    for key in list(updater.source):
        if not key.startswith('XS-Vcs-'):
            continue
        issue = LintianIssue(
            updater.source, 'adopted-extended-field', info=key)
        if not issue.should_fix():
            continue
        updater.source[key[3:]] = updater.source[key]
        del updater.source[key]
        issue.report_fixed()


report_result(
    "Remove unnecessary XS- prefix for Vcs- fields in debian/control.")



#!/usr/bin/python3

import sys
from email.utils import parseaddr

# Import convenience functions for reporting results and checking overrides
from lintian_brush.fixer import LintianIssue, control, report_result

PKG_PERL_EMAIL = 'pkg-perl-maintainers@lists.alioth.debian.org'
URL_BASE = 'https://salsa.debian.org/perl-team/modules/packages'

with control as e:
    # Parse the maintainer field and extract the email address.
    try:
        (name, email) = parseaddr(e.source['Maintainer'])
    except KeyError:
        sys.exit(2)
    if email != PKG_PERL_EMAIL:
        # Nothing to do here, it's not a pkg-perl-maintained package
        sys.exit(0)
    # Iterate over all fields in the source package
    for field in list(e.source):
        if not field.lower().startswith('vcs-'):
            # Ignore non-Vcs fields
            continue
        issue = LintianIssue(e.source, 'team/pkg-perl/vcs/no-git', info=field)
        if field.lower() not in ('vcs-git', 'vcs-browser'):
            if not issue.should_fix():
                continue
            # Drop this field
            del e.source[field]
            issue.report_fixed()

    for field, template in [
            ('Vcs-Git', URL_BASE + '/%s.git'),
            ('Vcs-Browser', URL_BASE + '/%s')]:
        old_value = e.source.get(field)
        issue = LintianIssue(
            e.source, 'team/pkg-perl/vcs/no-team-url',
            (field, old_value or ''))
        if not issue.should_fix():
            continue
        if old_value is not None and old_value.startswith(URL_BASE):
            continue

        e.source[field] = template % e.source['Source']
        # TODO(jelmer): Check that URLs actually exist, if net access is
        # allowed?
        issue.report_fixed()


report_result(
    'Use standard Vcs fields for perl package.',
    certainty='certain')



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.scrub_obsolete."""

from typing import Dict
from unittest import TestCase

from debmutate._deb822 import PkgRelation

from debian.changelog import Version
from lintian_brush.scrub_obsolete import (
    DropEssential,
    DropMinimumVersion,
    DropTransition,
    PackageChecker,
    ReplaceTransition,
    drop_obsolete_depends,
    filter_relations,
    name_list,
    release_aliases,
)


class NameListTests(TestCase):

    def test_empty(self):
        self.assertRaises(ValueError, name_list, [])

    def test_some(self):
        self.assertEqual('foo', name_list(['foo']))
        self.assertEqual('foo', name_list(['foo', 'foo', 'foo']))
        self.assertEqual(
            'bar and foo', name_list(['foo', 'bar', 'foo']))
        self.assertEqual(
            'bar, bla and foo', name_list(['foo', 'bar', 'foo', 'bla']))


class FilterRelationsTests(TestCase):

    def test_missing(self):
        control: Dict[str, str] = {}
        self.assertEqual(
            [],
            filter_relations(control, "Build-Depends", None))  # type: ignore

    def test_keep(self):
        control = {"Depends": "foo"}

        def cb(oldrel):
            return oldrel, []

        self.assertEqual([], filter_relations(control, "Depends", cb))

    def test_drop_last(self):
        control = {"Depends": "foo"}

        def cb(oldrel):
            return [], oldrel

        self.assertEqual(PkgRelation.parse("foo"),
                         filter_relations(control, "Depends", cb))
        self.assertEqual({}, control)

    def test_drop(self):
        control = {"Depends": "foo, bar"}

        def cb(oldrel):
            if oldrel[0].name == 'foo':
                return [], oldrel
            return oldrel, []

        self.assertEqual(PkgRelation.parse("foo"),
                         filter_relations(control, "Depends", cb))
        self.assertEqual({"Depends": "bar"}, control)

    def test_keep_last_comma(self):
        control = {"Depends": "foo, bar, "}

        def cb(oldrel):
            if oldrel and oldrel[0].name == 'foo':
                return [], oldrel
            return oldrel, []

        self.assertEqual(PkgRelation.parse("foo"),
                         filter_relations(control, "Depends", cb))
        self.assertEqual({"Depends": "bar, "}, control)

    def test_drop_just_comma(self):
        control = {"Depends": "foo, "}

        def cb(oldrel):
            if oldrel and oldrel[0].name == 'foo':
                return [], oldrel
            return oldrel, []

        self.assertEqual(PkgRelation.parse("foo"),
                         filter_relations(control, "Depends", cb))
        self.assertEqual({}, control)


class DummyChecker(PackageChecker):

    release = "release"

    def __init__(self, versions, essential, *, transitions=None):
        self.versions = versions
        self.essential = essential
        self._transitions = transitions or {}

    def package_version(self, package):
        return self.versions.get(package)

    def is_essential(self, package):
        return package in self.essential


class DropObsoleteDependsTests(TestCase):

    def test_empty(self):
        self.assertEqual(
            ([], []), drop_obsolete_depends([], DummyChecker({}, [])))

    def test_single(self):
        checker = DummyChecker({'simple': Version('1.1')}, {})
        orig = PkgRelation.parse('simple (>= 1.0)')
        self.assertEqual(
            (PkgRelation.parse('simple'),
             [DropMinimumVersion(*PkgRelation.parse('simple (>= 1.0)'))]),
            drop_obsolete_depends(orig, checker))

    def test_essential(self):
        checker = DummyChecker({'simple': Version('1.1')}, {'simple'})
        orig = PkgRelation.parse('simple (>= 1.0)')
        self.assertEqual(
            ([], [DropEssential(*PkgRelation.parse('simple (>= 1.0)'))]),
            drop_obsolete_depends(orig, checker))

    def test_debhelper(self):
        checker = DummyChecker({'debhelper': Version('1.4')}, {})
        orig = PkgRelation.parse('debhelper (>= 1.1)')
        self.assertEqual(
            (PkgRelation.parse('debhelper (>= 1.1)'), []),
            drop_obsolete_depends(orig, checker))

    def test_other_essential(self):
        checker = DummyChecker({'simple': Version('1.1')}, {'simple'})
        orig = PkgRelation.parse('simple (>= 1.0) | other')
        self.assertEqual(
            ([],
             [DropEssential(*PkgRelation.parse('simple (>= 1.0)'))]),
            drop_obsolete_depends(orig, checker))

    def test_transition(self):
        checker = DummyChecker({'simple': Version('1.1')}, {'simple'},
                               transitions={'oldpackage': 'replacement'})
        orig = PkgRelation.parse('oldpackage (>= 1.0) | other')
        self.assertEqual(
            (PkgRelation.parse('replacement | other'),
             [ReplaceTransition(PkgRelation.parse('oldpackage (>= 1.0)')[0],
                                PkgRelation.parse('replacement'))]),
            drop_obsolete_depends(orig, checker))

    def test_transition_matches(self):
        checker = DummyChecker({'simple': Version('1.1')}, {'simple'},
                               transitions={'oldpackage': 'replacement'})
        orig = PkgRelation.parse('oldpackage (>= 1.0) | replacement ')
        self.assertEqual(
            (PkgRelation.parse('replacement'),
             [DropTransition(*PkgRelation.parse('oldpackage (>= 1.0)'))]),
            drop_obsolete_depends(orig, checker))

    def test_transition_dupes(self):
        checker = DummyChecker({'simple': Version('1.1')}, {'simple'},
                               transitions={'oldpackage': 'replacement'})
        orig = PkgRelation.parse(
            'oldpackage (>= 1.0) | oldpackage (= 3.0) | other')
        self.assertEqual(
            (PkgRelation.parse('replacement | other'),
             [ReplaceTransition(PkgRelation.parse('oldpackage (>= 1.0)')[0],
                                PkgRelation.parse('replacement')),
              ReplaceTransition(PkgRelation.parse('oldpackage (= 3.0)')[0],
                                PkgRelation.parse('replacement'))]),
            drop_obsolete_depends(orig, checker))


class ReleaseAliasesTests(TestCase):

    def test_existing(self):
        self.assertEqual('(unstable)', release_aliases('sid'))

    def test_missing(self):
        self.assertEqual('', release_aliases('unknown'))



#!/usr/bin/python3

import os
from contextlib import suppress

from lintian_brush.fixer import LintianIssue, report_result

with suppress(FileNotFoundError, NotADirectoryError), \
        open('debian/tests/control') as f:
    if f.read().strip() == '':
        issue = LintianIssue('source', 'empty-debian-tests-control')
        if issue.should_fix():
            os.unlink('debian/tests/control')
            if not os.listdir('debian/tests'):
                os.rmdir('debian/tests')
            issue.report_fixed()

report_result(
    'Remove empty debian/tests/control.')



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, control, report_result

updated_packages = set()


with control as updater:
    for binary in updater.binaries:
        package = binary['Package']
        if (not package.startswith('fonts-') and
                not package.startswith('xfonts-')):
            continue
        if binary.get('Architecture') not in ('all', None):
            continue
        if 'Multi-Arch' in binary:
            continue
        issue = LintianIssue(
            updater.source, 'font-package-not-multi-arch-foreign')
        if issue.should_fix():
            binary['Multi-Arch'] = 'foreign'
            updated_packages.add(package)
            issue.report_fixed()


report_result(
    'Set Multi-Arch: foreign on package{} {}.'.format(
        's' if len(updated_packages) > 1 else '', ', '.join(updated_packages)))



#!/usr/bin/python3
# Report lintian tags that might be good candidates to implement fixers for.

import argparse

import psycopg2
from ruamel.yaml import YAML

from lintian_brush import available_lintian_fixers

parser = argparse.ArgumentParser()
parser.add_argument('--exclude', type=str, default='hard',
                    help='Comma-separated list of difficulties to exclude.')
args = parser.parse_args()

conn = psycopg2.connect(
    "postgresql://udd-mirror:udd-mirror@udd-mirror.debian.net/udd")
with conn.cursor() as cursor:
    cursor.execute(
        "SELECT tag, COUNT(DISTINCT package) AS package_count, "
        "COUNT(*) AS tag_count from lintian "
        "WHERE tag_type NOT IN ('classification') GROUP BY 1 ORDER BY 2 DESC")

    supported_tags = set()
    for fixer in available_lintian_fixers():
        supported_tags.update(fixer.lintian_tags)

    yaml = YAML()
    with open('tag-status.yaml') as f:
        tag_status = yaml.load(f)

    per_tag_status = {}
    for entry in tag_status or []:
        per_tag_status[entry['tag']] = entry

    exclude_difficulties = args.exclude.split(',')

    for (tag, package_count, tag_count) in cursor:
        if tag in supported_tags:
            continue
        difficulty = per_tag_status.get(tag, {}).get('difficulty', 'unknown')
        if difficulty in exclude_difficulties:
            continue
        print('%s %s %d/%d' % (tag, difficulty, package_count, tag_count))



#!/usr/bin/python3

import re

from lintian_brush.fixer import LintianIssue, report_result

with open('debian/copyright', 'rb') as f:
    lines = list(f)

m = re.match(
    rb'^(Format|Format-Specification): '
    rb'(http:\/\/www.debian.org\/doc\/packaging-manuals\/'
    rb'copyright-format\/1.0.*)\n', lines[0])
if m:
    newline = (
        b'Format: https://www.debian.org/doc/packaging-manuals/'
        b'copyright-format/1.0/\n')
    if newline != lines[0]:
        lines[0] = newline
        issue = LintianIssue(
            'source', 'insecure-copyright-format-uri', m.group(2).decode())
        if issue.should_fix():
            with open('debian/copyright', 'wb') as f:
                f.writelines(lines)
            issue.report_fixed()

report_result("Use secure copyright file specification URI.")



#!/usr/bin/python3

from urllib.parse import urlparse

from debmutate.watch import WatchEditor

from lintian_brush.fixer import (
    LintianIssue,
    current_package_version,
    net_access_allowed,
    report_result,
    source_package_name,
)
from lintian_brush.watch import watch_entries_certainty

try:
    with WatchEditor() as updater:
        changed_entries = []
        for w in getattr(updater.watch_file, 'entries', []):
            parsed_url = urlparse(w.url)
            if parsed_url.netloc != 'githubredir.debian.net':
                continue
            parts = parsed_url.path.strip('/').split('/')
            if parts[0] != 'github':
                # Hmm.
                continue
            issue = LintianIssue(
                'source', 'debian-watch-file-uses-deprecated-githubredir',
                info=f'{w.url} {w.matching_pattern}')
            if issue.should_fix():
                w.url = f'https://github.com/{parts[1]}/{parts[2]}/tags'
                w.matching_pattern = (
                    '.*/' + w.matching_pattern.rsplit('/', 1)[-1])
                issue.report_fixed()
                changed_entries.append(w)
        if net_access_allowed():
            certainty = watch_entries_certainty(
                changed_entries, source_package_name(),
                expected_versions=[current_package_version().upstream_version],
                default_certainty="confident")
        else:
            certainty = "confident"
except FileNotFoundError:
    pass
else:
    report_result(
        'Remove use of githubredir - see '
        'https://lists.debian.org/debian-devel-announce/2014/10/msg00000.html '
        'for details.', certainty=certainty)



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with lintian overrides files."""

import os
import re
from contextlib import suppress
from functools import partial
from typing import Callable, Iterator, List, Optional, Tuple

from debmutate.lintian_overrides import (
    LintianOverride,
    LintianOverridesEditor,
    iter_overrides,
)

from debian.changelog import Version

from . import open_binary


def overrides_paths() -> Iterator[str]:
    for path in ["debian/source/lintian-overrides"]:
        if os.path.exists(path):
            yield path
    if os.path.exists("debian"):
        for entry in os.scandir("debian"):
            if entry.name.endswith(".lintian-overrides"):
                yield entry.path


def update_overrides(
        cb: Callable[
            [str, int, LintianOverride],
            Optional[LintianOverride]]) -> None:
    """Call update_overrides_file on all overrides files.

    Args:
      cb: Callback that modifies overrides; called with path, linenumber and
         an Override object
    """
    for path in overrides_paths():
        update_overrides_file(partial(cb, path), path=path)


def update_overrides_file(
    cb: Callable[[int, LintianOverride], Optional[LintianOverride]],
    path: str = "debian/source/lintian-overrides",
) -> bool:
    """Modify the overrides in a file.

    Args:
      cb: Callback that modifies overrides; called with an Override object
        Should return new override or None to delete override.

    Returns:
        Whether the file was modified
    """
    with LintianOverridesEditor(path=path) as editor:
        new_lines = []
        comments: List[str] = []
        for lineno, entry in enumerate(editor.lines, 1):
            if isinstance(entry, LintianOverride):
                entry = cb(lineno, entry)
                if entry is not None:
                    new_lines.extend(comments)
                    new_lines.append(entry)
                comments = []
            else:
                comments.append(entry)
        if editor.lines != new_lines + comments:
            editor._parsed = new_lines + comments
            if editor._parsed == []:
                editor._parsed = None
        return editor.has_changed()


def get_overrides(
    type: Optional[str] = None, package: Optional[str] = None
) -> Iterator[LintianOverride]:
    paths = []
    if type in ("source", None):
        paths.extend(
            ["debian/source/lintian-overrides",
             "debian/source.lintian-overrides"]
        )
    if type in ("binary", None):
        if package is not None:
            paths.extend(["debian/%s.lintian-overrides" % package])
        elif os.path.isdir("debian"):
            paths.extend(
                [
                    e.path
                    for e in os.scandir("debian")
                    if e.name.endswith(".lintian-overrides")
                ]
            )

    for path in paths:
        with suppress(FileNotFoundError), open(path) as f:
            yield from iter_overrides(f)


def override_exists(
    tag: str,
    info: Optional[str] = None,
    package: Optional[str] = None,
    type: Optional[str] = None,
    arch: Optional[str] = None,
) -> bool:
    """Check if a particular override exists.

    Args:
      tag: Tag name
      info: Optional info
      package: Package
      type: package type (source, binary)
      arch: Architecture
    """
    return any(
        override.matches(package=package, info=info, tag=tag, arch=arch,
                         type=type)
        for override in get_overrides(type=type, package=package))


def get_unused_overrides(
    packages: List[Tuple[str, str]]
) -> List[Tuple[str, str, Version, str]]:
    from .udd import connect_udd_mirror

    args: List[str] = []
    extra = []
    for (type, name) in packages:
        extra.append(
            "package = $%d AND package_type = $%d" % (
                len(args) + 1, len(args) + 2)
        )
        args.extend([name, type])

    udd = connect_udd_mirror()
    with udd.cursor() as cursor:
        cursor.execute("""\
select package, package_type, package_version, information
from lintian where tag = 'unused-override' AND ({})
""".format(" OR ".join(extra), *args))
        return list(cursor)


unused_overrides = None


def remove_unused(
        control_paragraphs, ignore_tags=None) -> List[LintianOverride]:

    if ignore_tags is None:
        ignore_tags = set()
    packages = []
    for para in control_paragraphs:
        if "Source" in para:
            packages.append(("source", para["Source"]))
        else:
            packages.append(("binary", para["Package"]))
    global unused_overrides
    unused_overrides = None
    removed = []

    def is_unused(override, unused_overrides):
        for unused_override in unused_overrides:
            if override.package not in (None, unused_override[0]):
                continue
            if override.type not in (None, unused_override[1]):
                continue
            if override.info:
                expected_info = f"{override.tag} {override.info}"
            else:
                expected_info = override.tag
            if expected_info != unused_override[3]:
                continue
            return True
        return False

    def drop_override(path, lineno, override):
        global unused_overrides
        if unused_overrides is None:
            unused_overrides = get_unused_overrides(packages)
        if (is_unused(override, unused_overrides)
                and override.tag not in ignore_tags):
            removed.append(override)
            return None
        return override

    update_overrides(drop_override)
    return removed


def load_renamed_tags():
    import json
    with open_binary('renamed-tags.json') as f:
        return json.load(f)


LINENO_MATCH = r"(?P<lineno>\d+|\*)"
PATH_MATCH = r"(?P<path>(?!\\\[)[^[ ]+)"

# Common regexes, for convenience:

# "$file (line $lineno)" => "[$file:$lineno]"
PURE_FLN_SUB = (
    r"^" + PATH_MATCH + r" \(line " + LINENO_MATCH + r"\)$",
    r"[\1:\2]")
# "$file (line $lineno)" => "* [$file:$lineno]"
PURE_FLN_WILDCARD_SUB = (
    r"^" + PATH_MATCH + r" \(line " + LINENO_MATCH + r"\)$",
    r"* [\1:\2]")
# "$file (line $lineno) $msg" => "$msg [$file:$lineno]"
INTERTWINED_FLN_SUB = [
    (r"^" + PATH_MATCH + r" (.+) \(line (" + LINENO_MATCH + r"\)",
     r"\2 [\1:\3]"),
    (r"^" + PATH_MATCH + r" (.+) \*", r"\2 [\1:*]"),
    (r"^" + PATH_MATCH + r" \*", r"* [\1:*]"),
]
COPYRIGHT_SUB = [
    (r"^debian/copyright (.+) \(line " + LINENO_MATCH + r"\)",
     r"\1 [debian/copyright:\2]"),
    (r"^debian/copyright (.+) \*", r"\1 [debian/copyright:*]"),
    (r"^debian/copyright \*", r"* [debian/copyright:*]"),
    (r"^([^/ ]+) \*", r"\1 [debian/copyright:*]"),
]
# "$file" => "[$file]"
PURE_FN_SUB = (r"^" + PATH_MATCH + "$", r"[\1]")


# When adding new expressions here, make sure the first argument doesn't match
# on the new format.
INFO_FIXERS = {
    "autotools-pkg-config-macro-not-cross-compilation-safe":
        PURE_FLN_WILDCARD_SUB,
    "debian-rules-parses-dpkg-parsechangelog": PURE_FLN_SUB,
    "debian-rules-should-not-use-custom-compression-settings":
        (r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/rules:\2]"),
    "debian-source-options-has-custom-compression-settings":
        (r"(.*) \(line " + LINENO_MATCH + r"\)",
            r"\1 [debian/source/options:\2]"),
    "global-files-wildcard-not-first-paragraph-in-dep5-copyright":
        PURE_FLN_SUB,
    "missing-license-paragraph-in-dep5-copyright": COPYRIGHT_SUB,
    "missing-license-text-in-dep5-copyright": COPYRIGHT_SUB,
    "unused-license-paragraph-in-dep5-copyright": (
        r"([^ ]+) (.*) \(line " + LINENO_MATCH + r"\)",
        r"\2 [\1:\3]"),
    "license-problem-undefined-license": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/copyright:\2]"),
    "debhelper-tools-from-autotools-dev-are-deprecated": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/rules:\2]"),
    "version-substvar-for-external-package": (
        r"([^ ]+) \(line " + LINENO_MATCH + r"\) (.*)",
        r"\1 \3 [debian/control:\2]"),
    "debian-watch-uses-insecure-uri": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/watch:\2]"),
    "uses-deprecated-adttmp": (
        r"([^ ]+) \(line " + LINENO_MATCH + r"\)", r"[\1:\2]"),
    "incomplete-creative-commons-license": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/copyright:\2]"),
    "debian-rules-sets-dpkg-architecture-variable": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/rules:\2]"),
    "override_dh_auto_test-does-not-check-DEB_BUILD_OPTIONS": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/rules:\2]"),
    "dh-quilt-addon-but-quilt-source-format": (
        r"(.*) \(line " + LINENO_MATCH + r"\)", r"\1 [debian/rules:\2]"),
    "uses-dpkg-database-directly": PURE_FN_SUB,
    "package-contains-documentation-outside-usr-share-doc": PURE_FN_SUB,
    "non-standard-dir-perm": (
        r"^" + PATH_MATCH + r" ([0-9]+) \!= ([0-9]+)", r"\2 != \3 [\1]"),
    "non-standard-file-perm": (
        r"^" + PATH_MATCH + r" ([0-9]+) \!= ([0-9]+)", r"\2 != \3 [\1]"),
    "executable-is-not-world-readable": (
        r"^" + PATH_MATCH + r" ([0-9]+)", r"\2 [\1]"),
    "library-not-linked-against-libc": PURE_FN_SUB,
    "setuid-binary": (
        r"^" + PATH_MATCH + " (?P<mode>[0-9]+) (.+/.+)", r"\2 \3 [\1]"),
    "elevated-privileges": (
        r"^" + PATH_MATCH + " (?P<mode>[0-9]+) (.+/.+)", r"\2 \3 [\1]"),
    "executable-in-usr-lib": PURE_FN_SUB,
    "executable-not-elf-or-script": PURE_FN_SUB,
    "image-file-in-usr-lib": PURE_FN_SUB,
    "extra-license-file": PURE_FN_SUB,
    "script-not-executable": PURE_FN_SUB,
    "shell-script-fails-syntax-check": PURE_FN_SUB,
    "manpage-has-errors-from-man":
        (r"^" + PATH_MATCH + " ([^[]*)", r"\2 [\1]"),
    "groff-message": (
        r"^" + PATH_MATCH + " ([0-9]+): (.+)$", r"\2: \3 [\1:*]"),
    "source-contains-prebuilt-javascript-object": [
        PURE_FN_SUB,
        (r"^(?P<path>[^[ ].+) line length is .*", r"[\1]")],
    "source-contains-prebuilt-java-object": PURE_FN_SUB,
    "source-contains-prebuilt-windows-binary": PURE_FN_SUB,
    "source-contains-prebuilt-doxygen-documentation": PURE_FN_SUB,
    "source-contains-prebuilt-wasm-binary": PURE_FN_SUB,
    "source-contains-prebuilt-binary": PURE_FN_SUB,
    "source-is-missing": [
        (r"^(?P<path>[^[ ].+) line length is .*", r"[\1]"),
        (r"^(?P<path>[^[ ].+) \*", r"[\1]"),
        PURE_FN_SUB],
    "spelling-error-in-binary":
        (r"^" + PATH_MATCH + r" (.+) ([^[/\*]+)$", r"\2 \3 [\1]"),
    "very-long-line-length-in-source-file": [
        (PATH_MATCH +
         r" line ([0-9]+) is ([0-9]+) characters long \(>([0-9]+)\)",
         r"\3 > \4 [\1:\2]"),
        (PATH_MATCH +
         r" line length is ([0-9]+) characters \(>([0-9]+)\)",
         r"\2 > \3 [\1:*]"),
        (r"^" + PATH_MATCH + r" \*", r"* [\1:*]"),
        (r"^" + PATH_MATCH + r" line \*$", r"* [\1:*]")],
    "national-encoding": PURE_FN_SUB,
    "no-manual-page": PURE_FN_SUB,
    "package-contains-empty-directory": PURE_FN_SUB,
    "hardening-no-fortify-functions": PURE_FN_SUB,
    "maintainer-manual-page": PURE_FN_SUB,
    "library-package-name-for-application": PURE_FN_SUB,
    "script-with-language-extension": PURE_FN_SUB,
    "license-problem-non-free-img-lenna": PURE_FN_SUB,
    "file-without-copyright-information": ("^(.*)$", r"\1 [debian/copyright]"),
    "globbing-patterns-out-of-order":
        ("^(?P<path>.+) ([^[]+)$", r"\2 [\1:*]"),
    "statically-linked-binary": PURE_FN_SUB,
    "spare-manual-page": PURE_FN_SUB,
    "shared-library-lacks-prerequisites": PURE_FN_SUB,
    "codeless-jar": PURE_FN_SUB,
    "hardening-no-pie": PURE_FN_SUB,
    "hardening-no-relro": PURE_FN_SUB,
    "obsolete-url-in-packaging": (
        r"^(?P<path>.+) (.+)$", r"\2 [\1]"),
    "inconsistent-appstream-metadata-license": (
        r'^(.+) \(([^ ]+) \!= ([^)]+)\)$',
        r'\1 (\2 != \3) [debian/copyright]'),
    "source-ships-excluded-file": (
        r"^([^ ]+)$", r"\1 [debian/copyright:*]"),
    "package-installs-java-bytecode": PURE_FN_SUB,
    "jar-not-in-usr-share": PURE_FN_SUB,
    "debconf-is-not-a-registry": ("^" + PATH_MATCH + "$", r"[\1:*]"),
    "unused-debconf-template": ("^([^ ]+)$", r"\1 [*:*]"),
    "apache2-reverse-dependency-calls-invoke-rc.d": (
        "^" + PATH_MATCH + r":([0-9]+)$", r"[\1:\2]"),
    "application-in-library-section": (
        "^([^ ]+) " + PATH_MATCH + "$", r"\1 [\2]"),
    "repeated-path-segment": (
        "^([^ ]+) " + PATH_MATCH + "$", r"\1 [\2]"),
    "symlink-is-self-recursive": (
        "^([^ ]+) " + PATH_MATCH + "$", r"\1 [\2]"),
    "privacy-breach-google-adsense": (
        r"^" + PATH_MATCH + r" \(choke on: ([^\)]+)\)$",
        r"(choke on: \2) [\1]"),
    "systemd-service-file-refers-to-unusual-wantedby-target": (
        "^" + PATH_MATCH + r" ([^[ ]+)$", r"\2 [\1]"),
    "duplicate-font-file": (
        "^" + PATH_MATCH + r" also in ([^[]+)$",
        r"also in (\2) [\1]"),
}


def fix_override_info(override):
    try:
        fixers = INFO_FIXERS[override.tag]
    except KeyError:
        # no regex available
        return override.info
    else:
        info = override.info
        if not isinstance(fixers, list):
            fixers = [fixers]  # type: ignore
        for fixer in fixers:
            if isinstance(fixer, tuple):
                info = re.sub(fixer[0], fixer[1], info)
                # The regex should only apply once
                if re.sub(fixer[0], fixer[1], info) != info:
                    raise AssertionError(
                        "invalid repeatable regex for {}: {}".format(
                             override.tag, fixer[0]))
            elif callable(fixer):
                info = fixer(info) or info
            else:
                raise TypeError(fixer)
        return info


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--remove-unused", action="store_true", help="Remove unused overrides."
    )
    args = parser.parse_args()
    if args.remove_unused:
        from debian.deb822 import Deb822
        with open("debian/control") as f:
            removed = remove_unused(Deb822.iter_paragraphs(f))
        print("Removed %d unused overrides" % len(removed))
    else:
        parser.print_usage()



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.lintian_overrides import update_overrides

REMOVED_TAGS = [
    'hardening-no-stackprotector',
    'maintainer-not-full-name',
    'uploader-not-full-name',
    'uploader-address-missing',
    'no-upstream-changelog',
    'copyright-year-in-future',
    'script-calls-init-script-directly',
    ]

# TODO(jelmer): Check if a tag matches a binary package name.


removed = []


def fix_malformed(path, lineno, override):
    if override.tag not in REMOVED_TAGS:
        return override
    issue = LintianIssue(
        (override.type, override.package), 'malformed-override',
        'Unknown tag %s in line %d' % (override.tag, lineno))
    if issue.should_fix():
        issue.report_fixed()
        removed.append(override.tag)
        return None
    return override


update_overrides(fix_malformed)

report_result(
    'Remove overrides for lintian tags that are no longer supported: %s' %
    ', '.join(removed))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import shlex
import subprocess
import tempfile

gpg = shlex.split(os.environ.get("GPG", "gpg"))

KEY_BLOCK_START = b"-----BEGIN PGP PUBLIC KEY BLOCK-----"
KEY_BLOCK_END = b"-----END PGP PUBLIC KEY BLOCK-----"


class GpgMissing(Exception):
    """gpg command missing."""


class GpgFailed(Exception):
    """gpg command failed."""


def gpg_import_export(import_options, export_options, stdin):
    argv = gpg + [
        "--armor",
        "--quiet",
        "--no-default-keyring",
        "--export-options",
        ",".join(export_options),
        "--import-options",
        ",".join(["import-export"] + import_options),
        "--output",
        "-",
        "--import",
        "-",
    ]
    with tempfile.TemporaryDirectory() as td:
        try:
            p = subprocess.Popen(
                argv, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                env={'GNUPGHOME': td})
        except FileNotFoundError as exc:
            # No gpg, no dice.
            raise GpgMissing() from exc
        (stdout, stderr) = p.communicate(stdin, timeout=5)
        if p.returncode != 0:
            raise GpgFailed(stderr)
        return stdout


def fetch_keys(keys, home_dir):
    import subprocess
    env = dict(os.environ)
    if home_dir:
        env['GNUPGHOME'] = home_dir
    try:
        subprocess.check_call(
            ['gpg', '--recv-keys'] + keys,
            env=env)
    except FileNotFoundError as exc:
        # No gpg, no dice.
        raise GpgMissing() from exc
    except subprocess.CalledProcessError:
        return False
    return True



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import LintianIssue, report_result

try:
    with open('debian/source/options') as f:
        oldlines = list(f.readlines())
except FileNotFoundError:
    sys.exit(0)


def drop_prior_comments(lines):
    while lines and lines[-1].strip().startswith('#'):
        lines.pop(-1)


dropped = set()

newlines = []
for lineno, line in enumerate(oldlines, 1):
    if line.lstrip().startswith('#'):
        newlines.append(line)
        continue
    try:
        (key, value) = line.split('=', 1)
    except ValueError:
        newlines.append(line)
    else:
        if key.strip() == 'compression':
            issue = LintianIssue(
                'source', 'custom-compression-in-debian-source-options',
                '%s (line %d)' % (line.rstrip('\n'), lineno))
            if issue.should_fix():
                drop_prior_comments(newlines)
                dropped.add("custom source compression")
                issue.report_fixed()
                continue
        if key.strip() == 'compression-level':
            issue = LintianIssue(
                'source', 'custom-compression-in-debian-source-options',
                '%s (line %d)' % (line.rstrip('\n'), lineno))
            if issue.should_fix():
                drop_prior_comments(newlines)
                dropped.add("custom source compression level")
                issue.report_fixed()
                continue
        newlines.append(line)

if newlines:
    with open('debian/source/options', 'w') as f:
        f.writelines(newlines)
elif dropped:
    os.unlink('debian/source/options')

report_result('Drop %s.' % ', '.join(sorted(dropped)))



#!/usr/bin/python3

import re
import sys

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from lintian_brush.fixer import LintianIssue, report_result

used = []
defined = set()
certainty = 'certain'


def extract_licenses(synopsis):
    """Extract license names from a synopsis.

    This will return a list of licenses, as a list of possible names per
    license.
    """
    ret = []
    for license in synopsis.split(" or "):
        options = [license]
        m = re.fullmatch(r'(.*) with (.*) exception', license)
        if m:
            license = m.group(1)
        options.append(license)
        ret.append(options)
    return ret


try:  # noqa: C901
    with CopyrightEditor() as updater:
        if (updater.copyright.header.license
                and updater.copyright.header.license.text):
            defined.add(updater.copyright.header.license.synopsis)
        for paragraph in updater.copyright.all_paragraphs():
            if not paragraph.license:
                continue
            if paragraph.license.text:
                defined.add(paragraph.license.synopsis)

        if updater.copyright.header.license:
            synopsis = updater.copyright.header.license.synopsis
            if synopsis:
                if synopsis in defined:
                    used.append([synopsis])
                used.extend(extract_licenses(synopsis))

        for paragraph in updater.copyright.all_files_paragraphs():
            if not paragraph.license:
                continue
            if paragraph.files:
                synopsis = paragraph.license.synopsis
                if synopsis in defined:
                    used.append([synopsis])
                used.extend(extract_licenses(synopsis))

        extra_defined = set(defined)
        for options in used:
            extra_defined -= set(options)

        extra_used = []
        for options in used:
            for option in options:
                if option in defined:
                    break
            else:
                extra_used.append(options)

        if extra_used:
            sys.stderr.write('Undefined licenses in copyright: %r' %
                             [options[0] for options in extra_used])
            # Drop the certainty since it's possible the undefined licenses
            # are actually the referenced ones.
            certainty = 'possible'

        for name in extra_defined:
            for paragraph in updater.copyright.all_paragraphs():
                if not paragraph.license:
                    continue
                if paragraph.license.synopsis == name:
                    continue
                if paragraph.license.text and name in paragraph.license.text:
                    certainty = 'possible'
                if paragraph.comment and name in paragraph.comment:
                    certainty = 'possible'

        if extra_defined and not extra_used:
            for paragraph in list(updater.copyright.all_paragraphs()):
                if not paragraph.license:
                    continue
                issue = LintianIssue(
                    'source', 'unused-license-paragraph-in-dep5-copyright',
                    info=paragraph.license.synopsis.lower())
                if not issue.should_fix():
                    continue
                if paragraph.license.synopsis in extra_defined:
                    issue.report_fixed()
                    updater.remove(paragraph)
except (FileNotFoundError, NotMachineReadableError):
    pass
else:
    report_result(
        "Remove unused license definitions for %s." % ', '.join(extra_defined),
        certainty=certainty)



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    control,
    report_result,
)
from lintian_brush.lintian import LINTIAN_DATA_PATH


def parse(t):
    return tuple([int(v) for v in t.split(".")])


PYTHON_VERSIONS_PATH = os.path.join(
    LINTIAN_DATA_PATH, 'python/versions')


python_versions = {}
try:
    with open(os.path.join(PYTHON_VERSIONS_PATH)) as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            key, value = line.split('=')
            python_versions[key.strip()] = parse(value.strip())
except FileNotFoundError:
    sys.exit(2)


with control as updater:
    # Remove anything that involves python 2.6, 2.7, 3.3
    if ("X-Python-Version" in updater.source and
            updater.source["X-Python-Version"].strip().startswith(">= 2.")):
        vers = updater.source["X-Python-Version"].split(">=")[1].strip()
        if parse(vers) <= python_versions['old-python2']:
            if parse(vers) <= python_versions['ancient-python2']:
                kind = 'ancient'
            else:
                kind = 'old'
        issue = LintianIssue(
            updater.source, '%s-python-version-field' % kind,
            'x-python-version %s' % updater.source['X-Python-Version'])
        if issue.should_fix():
            del updater.source["X-Python-Version"]
            issue.report_fixed()
    if ("X-Python3-Version" in updater.source and
            updater.source["X-Python3-Version"].strip().startswith(">=")):
        vers = updater.source["X-Python3-Version"].split(">=")[1].strip()
        if parse(vers) <= python_versions['old-python3']:
            if parse(vers) <= python_versions['ancient-python3']:
                kind = 'ancient'
            else:
                kind = 'old'
            issue = LintianIssue(
                updater.source, '%s-python-version-field' % kind,
                'x-python3-version %s' % updater.source['X-Python3-Version'])
            if issue.should_fix():
                del updater.source["X-Python3-Version"]
                issue.report_fixed()


report_result(
    "Remove unnecessary X-Python{,3}-Version field in debian/control.")



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the vcs module."""

from unittest import TestCase

from lintian_brush.vcs import (
    canonicalize_vcs_url,
    determine_browser_url,
    find_public_vcs_url,
    fixup_broken_git_url,
    fixup_rcp_style_git_url,
    sanitize_url,
)


class FixUpGitUrlTests(TestCase):
    def test_fixup(self):
        self.assertEqual(
            "https://github.com/jelmer/dulwich",
            fixup_broken_git_url("https://github.com:jelmer/dulwich"),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich -b blah",
            fixup_broken_git_url("https://github.com:jelmer/dulwich -b blah"),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich",
            fixup_broken_git_url("git://github.com/jelmer/dulwich"),
        )

    def test_preserves(self):
        self.assertEqual(
            "https://github.com/jelmer/dulwich",
            fixup_broken_git_url("https://github.com/jelmer/dulwich"),
        )

    def test_salsa_not_https(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/dulwich",
            fixup_broken_git_url("git://salsa.debian.org/jelmer/dulwich"),
        )

    def test_salsa_uses_cgit(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/dulwich",
            fixup_broken_git_url(
                "https://salsa.debian.org/cgit/jelmer/dulwich"),
        )

    def test_salsa_tree_branch(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/dulwich -b master",
            fixup_broken_git_url(
                "https://salsa.debian.org/jelmer/dulwich/tree/master"),
        )

    def test_strip_extra_slash(self):
        self.assertEqual(
            "https://salsa.debian.org/salve/auctex.git",
            fixup_broken_git_url("https://salsa.debian.org//salve/auctex.git"),
        )

    def test_strip_extra_colon(self):
        self.assertEqual(
            "https://salsa.debian.org/mckinstry/lcov.git",
            fixup_broken_git_url(
                "https://salsa.debian.org:/mckinstry/lcov.git"),
        )

    def test_strip_username(self):
        self.assertEqual(
            "https://github.com/RPi-Distro/pgzero.git",
            fixup_broken_git_url(
                "git://git@github.com:RPi-Distro/pgzero.git"),
        )
        self.assertEqual(
            "https://salsa.debian.org/debian-astro-team/pyavm.git",
            fixup_broken_git_url(
                "https://git@salsa.debian.org:debian-astro-team/pyavm.git"
            ),
        )

    def test_github_tree_url(self):
        self.assertEqual(
            "https://github.com/blah/blah -b master",
            fixup_broken_git_url("https://github.com/blah/blah/tree/master"),
        )

    def test_freedesktop(self):
        self.assertEqual(
            "https://gitlab.freedesktop.org/xorg/xserver",
            fixup_broken_git_url("git://anongit.freedesktop.org/xorg/xserver"),
        )
        self.assertEqual(
            "https://gitlab.freedesktop.org/xorg/lib/libSM",
            fixup_broken_git_url(
                "git://anongit.freedesktop.org/git/xorg/lib/libSM"),
        )

    def test_anongit(self):
        self.assertEqual(
            "https://anongit.kde.org/kdev-php.git",
            fixup_broken_git_url("git://anongit.kde.org/kdev-php.git"),
        )

    def test_gnome(self):
        self.assertEqual(
            "https://gitlab.gnome.org/GNOME/alacarte",
            fixup_broken_git_url("https://git.gnome.org/browse/alacarte"),
        )


class SanitizeUrlTests(TestCase):
    def test_simple(self):
        self.assertEqual(
            "https://github.com/jelmer/blah.git",
            sanitize_url("http://github.com/jelmer/blah"),
        )

    def test_git_http(self):
        self.assertEqual(
            "https://github.com/jelmer/blah.git",
            sanitize_url("git+http://github.com/jelmer/blah"),
        )
        self.assertEqual(
            "https://github.com/jelmer/blah.git",
            sanitize_url("git+https://github.com/jelmer/blah"),
        )

    def test_rcp_style(self):
        self.assertEqual(
            "https://github.com/jelmer/blah.git",
            sanitize_url("github.com:jelmer/blah")
        )

    def test_insecure_savannah(self):
        self.assertEqual(
            "https://git.savannah.gnu.org/cgit/gforth.git",
            sanitize_url("http://git.savannah.gnu.org/cgit/gforth.git"),
        )

    def test_cvs(self):
        self.assertEqual(
            "cvs+ssh://_anoncvs@anoncvs.mirbsd.org/cvs#jupp",
            sanitize_url([":extssh:_anoncvs@anoncvs.mirbsd.org:/cvs", "jupp"]),
        )
        self.assertEqual(
            "cvs+pserver://_anoncvs@anoncvs.mirbsd.org/cvs#jupp",
            sanitize_url(
                [":pserver:_anoncvs@anoncvs.mirbsd.org:/cvs", "jupp"]),
        )


class DetermineBrowserUrlTests(TestCase):
    def test_salsa(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/dulwich",
            determine_browser_url(
                "git", "https://salsa.debian.org/jelmer/dulwich.git"),
        )

    def test_github(self):
        self.assertEqual(
            "https://github.com/jelmer/dulwich",
            determine_browser_url(
                "git", "https://github.com/jelmer/dulwich.git"),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich/tree/master",
            determine_browser_url(
                "git", "https://github.com/jelmer/dulwich.git -b master"
            ),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich/tree/master",
            determine_browser_url(
                "git", "git://github.com/jelmer/dulwich -b master"),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich/tree/master/blah",
            determine_browser_url(
                "git", "git://github.com/jelmer/dulwich -b master [blah]"
            ),
        )
        self.assertEqual(
            "https://github.com/jelmer/dulwich/tree/HEAD/blah",
            determine_browser_url(
                "git", "git://github.com/jelmer/dulwich [blah]"),
        )
        self.assertEqual(
            "https://git.sv.gnu.org/cgit/rcs.git",
            determine_browser_url("git", "https://git.sv.gnu.org/git/rcs.git"),
        )
        self.assertEqual(
            "https://git.savannah.gnu.org/cgit/rcs.git",
            determine_browser_url("git", "git://git.savannah.gnu.org/rcs.git"),
        )
        self.assertEqual(
            "https://sourceforge.net/p/shorewall/debian",
            determine_browser_url(
                "git", "git://git.code.sf.net/p/shorewall/debian"),
        )
        self.assertEqual(
            "https://sourceforge.net/p/shorewall/debian/ci/foo/tree",
            determine_browser_url(
                "git", "git://git.code.sf.net/p/shorewall/debian -b foo"
            ),
        )
        self.assertEqual(
            "https://sourceforge.net/p/shorewall/debian/ci/HEAD/tree/sp",
            determine_browser_url(
                "git", "git://git.code.sf.net/p/shorewall/debian [sp]"
            ),
        )
        self.assertEqual(
            "https://sourceforge.net/p/shorewall/debian/ci/foo/tree/sp",
            determine_browser_url(
                "git", "git://git.code.sf.net/p/shorewall/debian -b foo [sp]"
            ),
        )


class CanonicalizeVcsUrlTests(TestCase):
    def test_github(self):
        self.assertEqual(
            "https://github.com/jelmer/example.git",
            canonicalize_vcs_url("Git", "https://github.com/jelmer/example"),
        )

    def test_salsa(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/example.git",
            canonicalize_vcs_url(
                "Git", "https://salsa.debian.org/jelmer/example"),
        )
        self.assertEqual(
            "https://salsa.debian.org/jelmer/example.git",
            canonicalize_vcs_url(
                "Git", "https://salsa.debian.org/jelmer/example.git"),
        )


class FindPublicVcsUrlTests(TestCase):
    def test_github(self):
        self.assertEqual(
            "https://github.com/jelmer/example",
            find_public_vcs_url("ssh://git@github.com/jelmer/example"),
        )
        self.assertEqual(
            "https://github.com/jelmer/example",
            find_public_vcs_url("https://github.com/jelmer/example"),
        )

    def test_salsa(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/example",
            find_public_vcs_url("ssh://salsa.debian.org/jelmer/example"),
        )
        self.assertEqual(
            "https://salsa.debian.org/jelmer/example",
            find_public_vcs_url("https://salsa.debian.org/jelmer/example"),
        )


class FixupRcpStyleUrlTests(TestCase):
    def test_fixup(self):
        self.assertEqual(
            "ssh://github.com/jelmer/example",
            fixup_rcp_style_git_url("github.com:jelmer/example"),
        )
        self.assertEqual(
            "ssh://git@github.com/jelmer/example",
            fixup_rcp_style_git_url("git@github.com:jelmer/example"),
        )

    def test_leave(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/example",
            fixup_rcp_style_git_url("https://salsa.debian.org/jelmer/example"),
        )
        self.assertEqual(
            "ssh://git@salsa.debian.org/jelmer/example",
            fixup_rcp_style_git_url(
                "ssh://git@salsa.debian.org/jelmer/example"),
        )



#!/usr/bin/python3

import os
from contextlib import suppress

from lintian_brush.fixer import report_result

with suppress(FileNotFoundError), open('debian/source/options') as f:
    if not f.read().strip():
        os.unlink('debian/source/options')

report_result('Remove empty debian/source/options.')



#!/usr/bin/python3

import sys

from debmutate.control import (
    guess_template_type,
)
from debmutate.reformatting import (
    GeneratedFile,
    check_generated_file,
)

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor


def rewrite_line(line):
    if not line.split(b'#', 1)[0].strip():
        return line
    if line.startswith(b' ') or line.startswith(b'\t'):
        return line
    (key, value) = line.split(b':', 1)
    if not value.strip():
        return line
    return b'%s: %s\n' % (key, value.lstrip().rstrip(b'\n'))


def fix_field_spacing(path):
    changed = False
    with LineEditor(path, 'b') as e:
        for lineno, oldline in e:
            newline = rewrite_line(oldline)
            if newline != oldline:
                if path == 'debian/control':
                    issue = LintianIssue(
                        'source', 'debian-control-has-unusual-field-spacing',
                        info='line %d' % lineno)
                    if issue.should_fix():
                        e[lineno] = newline
                        changed = True
                        issue.report_fixed()
                else:
                    e[lineno] = newline
                    changed = True
    return changed


try:
    check_generated_file('debian/control')
except GeneratedFile as e:
    if e.template_path:
        template_type = guess_template_type(e.template_path, 'debian')
        if template_type is None:
            raise
        changed = fix_field_spacing(e.template_path)
        if changed:
            fix_field_spacing('debian/control')
    else:
        raise
except FileNotFoundError:
    sys.exit(0)
else:
    try:
        changed = fix_field_spacing('debian/control')
    except FileNotFoundError:
        sys.exit(0)

if changed:
    report_result('Strip unusual field spacing from debian/control.')



#!/usr/bin/python3

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.systemd import (
    SystemdServiceEditor,
    Undefined,
    systemd_service_files,
)

for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        old_pidfile = updater.file['Service']['PIDFile']
        if isinstance(old_pidfile, str):
            new_pidfile = old_pidfile.replace("/var/run/", "/run/")
            updater.file['Service']['PIDFile'] = new_pidfile
            fixed_lintian_tag(
                'source',
                'systemd-service-file-refers-to-var-run',
                (path, 'PIDFile', old_pidfile))
        for key in updater.file['Service']:
            val = updater.file['Service'][key]
            if isinstance(old_pidfile, Undefined) or old_pidfile not in val:
                continue
            updater.file['Service'][key] = val.replace(
                old_pidfile, new_pidfile)

report_result("Replace /var/run with /run for the Service PIDFile.")



#!/usr/bin/python3

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.systemd import SystemdServiceEditor, systemd_service_files

for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        unit = updater.file['Unit']
        try:
            unit.rename('BindTo', 'BindsTo')
        except KeyError:
            pass
        else:
            fixed_lintian_tag(
                'source', 'systemd-service-file-refers-to-obsolete-bindto',
                path)


report_result('Rename BindTo key to BindsTo in systemd files.')



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for accessing the DebBugs database."""

import logging


class DebBugs:
    """Read DebBugs data through UDD."""

    def __init__(self):
        self._conn = None

    def connect(self):
        from .udd import connect_udd_mirror

        self._conn = connect_udd_mirror()

    def check_bug(self, package, bugid):
        """Check that a bug belongs to a particular package.

        Args:
          package: Package name
          bugid: Bug number
        Returns:
          Boolean
        """
        assert self._conn is not None, "call connect() first"
        with self._conn.cursor() as cursor:
            cursor.execute("select package from bugs where id = %s", (bugid, ))
            row = cursor.fetchone()
            if row is None:
                return False
            return row[0] == package


def find_archived_wnpp_bugs(source_name):
    try:
        from .udd import connect_udd_mirror
    except ModuleNotFoundError:
        logging.warning("psycopg2 not available, unable to find wnpp bugs.")
        return []
    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute("""
select id, substring(title, 0, 3) from archived_bugs where package = 'wnpp' and
title like 'ITP: ' || %s || ' -- %%' OR
title like 'RFP: ' || %s || ' -- %%'
""", (source_name, source_name))
        return [(row[0], row[1]) for row in cursor]


def find_wnpp_bugs(source_name):
    try:
        from .udd import connect_udd_mirror
    except ModuleNotFoundError:
        logging.warning("psycopg2 not available, unable to find wnpp bugs.")
        return []
    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute("""
select id, type from wnpp where source = %s and type in ('ITP', 'RFP')
""", (source_name, ))
        return [(row[0], row[1]) for row in cursor]
